#!/usr/bin/python

# Copyright (C) 2006  Branko Kokanovic
#
#   DlgAdvOpt.py: dialog for advanced interpolation options for director
#

import gi

gi.require_version('Gtk', '3.0')
from gi.repository import Gtk


class DlgAdvOptions:
    def __init__(self, current_kf, animation, parent):
        # pylint: disable=no-member
        self.dialog = Gtk.Dialog(
            transient_for=parent,
            title="Keyframe advanced options",
            modal=True,
            destroy_with_parent=True
        )
        self.dialog.add_buttons(Gtk.STOCK_OK, Gtk.ResponseType.OK,
                                Gtk.STOCK_CANCEL, Gtk.ResponseType.CANCEL)

        self.current_kf = current_kf
        self.animation = animation

        tbl_main = Gtk.Grid()
        tbl_main.set_row_spacing(10)
        tbl_main.set_column_spacing(10)
        tbl_main.set_property("margin", 10)

        dirs = animation.get_directions(self.current_kf)

        lbl_xy = Gtk.Label(label="XY angles interpolation direction:")
        tbl_main.attach(lbl_xy, 0, 0, 1, 1)
        self.cmb_xy = Gtk.ComboBoxText()
        self.cmb_xy.append_text("Nearer")
        self.cmb_xy.append_text("Longer")
        self.cmb_xy.append_text("Clockwise")
        self.cmb_xy.append_text("Counterclockwise")
        self.cmb_xy.set_active(dirs[0])
        tbl_main.attach(self.cmb_xy, 1, 0, 1, 1)
        lbl_xz = Gtk.Label(label="XZ angles interpolation direction:")
        tbl_main.attach(lbl_xz, 0, 1, 1, 1)
        self.cmb_xz = Gtk.ComboBoxText()
        self.cmb_xz.append_text("Nearer")
        self.cmb_xz.append_text("Longer")
        self.cmb_xz.append_text("Clockwise")
        self.cmb_xz.append_text("Counterclockwise")
        self.cmb_xz.set_active(dirs[1])
        tbl_main.attach(self.cmb_xz, 1, 1, 1, 1)
        lbl_xw = Gtk.Label(label="XW angles interpolation direction:")
        tbl_main.attach(lbl_xw, 0, 2, 1, 1)
        self.cmb_xw = Gtk.ComboBoxText()
        self.cmb_xw.append_text("Nearer")
        self.cmb_xw.append_text("Longer")
        self.cmb_xw.append_text("Clockwise")
        self.cmb_xw.append_text("Counterclockwise")
        self.cmb_xw.set_active(dirs[2])
        tbl_main.attach(self.cmb_xw, 1, 2, 1, 1)
        lbl_yz = Gtk.Label(label="YZ angles interpolation direction:")
        tbl_main.attach(lbl_yz, 0, 3, 1, 1)
        self.cmb_yz = Gtk.ComboBoxText()
        self.cmb_yz.append_text("Nearer")
        self.cmb_yz.append_text("Longer")
        self.cmb_yz.append_text("Clockwise")
        self.cmb_yz.append_text("Counterclockwise")
        self.cmb_yz.set_active(dirs[3])
        tbl_main.attach(self.cmb_yz, 1, 3, 1, 1)
        lbl_yw = Gtk.Label(label="YW angles interpolation direction:")
        tbl_main.attach(lbl_yw, 0, 4, 1, 1)
        self.cmb_yw = Gtk.ComboBoxText()
        self.cmb_yw.append_text("Nearer")
        self.cmb_yw.append_text("Longer")
        self.cmb_yw.append_text("Clockwise")
        self.cmb_yw.append_text("Counterclockwise")
        self.cmb_yw.set_active(dirs[4])
        tbl_main.attach(self.cmb_yw, 1, 4, 1, 1)
        lbl_zw = Gtk.Label(label="ZW angles interpolation direction:")
        tbl_main.attach(lbl_zw, 0, 5, 1, 1)
        self.cmb_zw = Gtk.ComboBoxText()
        self.cmb_zw.append_text("Nearer")
        self.cmb_zw.append_text("Longer")
        self.cmb_zw.append_text("Clockwise")
        self.cmb_zw.append_text("Counterclockwise")
        self.cmb_zw.set_active(dirs[5])
        tbl_main.attach(self.cmb_zw, 1, 5, 1, 1)
        self.dialog.vbox.pack_start(tbl_main, True, True, 0)

    def show(self):
        self.dialog.show_all()
        response = self.dialog.run()
        if response == Gtk.ResponseType.OK:
            dirs = (self.cmb_xy.get_active(), self.cmb_xz.get_active(),
                    self.cmb_xw.get_active(), self.cmb_yz.get_active(),
                    self.cmb_yw.get_active(), self.cmb_zw.get_active())
            self.animation.set_directions(self.current_kf, dirs)

        self.dialog.destroy()
        return



#!/usr/bin/env python3

import copy
import io
import math
import random
from time import time as now

from fract4d_compiler import fc, fctutils
from . import colorizer, formsettings, fract4dc, gradient

# the version of the earliest gf4d release which can parse all the files
# this version can output
THIS_FORMAT_VERSION = "3.10"

BLEND_NEAREST = 0
BLEND_FURTHEST = 1
BLEND_CW = 2
BLEND_CCW = 3


class T(fctutils.T):
    XCENTER = 0
    YCENTER = 1
    ZCENTER = 2
    WCENTER = 3
    MAGNITUDE = 4
    XYANGLE = 5
    XZANGLE = 6
    XWANGLE = 7
    YZANGLE = 8
    YWANGLE = 9
    ZWANGLE = 10

    FORMULA = 0
    OUTER = 1
    INNER = 2
    DEFAULT_FORMULA_FILE = "gf4d.frm"
    DEFAULT_FORMULA_FUNC = "Mandelbrot"
    paramnames = [
        "x",
        "y",
        "z",
        "w",
        "size",
        "xy",
        "xz",
        "xw",
        "yz",
        "yw",
        "zw"]

    def __init__(self, compiler, site=None):
        fctutils.T.__init__(self)

        self.format_version = 2.8

        self.bailfunc = 0
        # formula support
        self.forms = [
            formsettings.T(compiler, self),  # formula
            formsettings.T(compiler, self, "cf0"),  # outer
            formsettings.T(compiler, self, "cf1")  # inner
        ]

        self.transforms = []
        self.next_transform_id = 0
        self.compiler_options = {"optimize": 1}
        self.yflip = False
        self.periodicity = True
        self.period_tolerance = 1.0E-9
        self.auto_epsilon = False  # automatically set @epsilon param, if found
        self.auto_deepen = True  # automatically adjust maxiter
        self.auto_tolerance = True  # automatically adjust periodicity
        self.antialias = 1
        self.compiler = compiler
        self.outputfile = None
        self.render_type = 0
        self.pfunc = None
        self.handle = None

        self.warp_param = None
        # gradient

        # default is just white outside
        self.default_gradient = gradient.Gradient()
        self.default_gradient.segments[0].left_color = [1.0, 1.0, 1.0, 1.0]
        self.default_gradient.segments[0].right_color = [1.0, 1.0, 1.0, 1.0]

        # solid colors are black
        self.solids = [(0, 0, 0, 255), (0, 0, 0, 255)]

        # formula defaults
        self.set_formula(T.DEFAULT_FORMULA_FILE, T.DEFAULT_FORMULA_FUNC, 0)
        self.set_inner("gf4d.cfrm", "zero")
        self.set_outer("gf4d.cfrm", "continuous_potential")
        self.dirtyFormula = True  # formula needs recompiling
        self.dirty = True  # parameters have changed
        self.clear_image = True

        self.reset()

        # interaction with fract4dc library
        self.site = site or fract4dc.site_create(self)

        # colorfunc lookup
        self.colorfunc_names = [
            "default",
            "continuous_potential",
            "zero",
            "ejection_distance",
            "decomposition",
            "external_angle"]

        self.saved = True  # initial params not worth saving

    def serialize(self, comp=False):
        out = io.StringIO()
        self.save(out, False, compress=comp)
        return out.getvalue()

    def deserialize(self, string):
        self.loadFctFile(io.StringIO(string))
        self.changed()

    def apply_params(self, dict):
        for (key, value) in list(dict.items()):
            self.parseVal(key, value, None)

    def save(self, file, update_saved_flag=True, **kwds):
        print("gnofract4d parameter file", file=file)
        print("version=%s" % THIS_FORMAT_VERSION, file=file)

        compress = kwds.get("compress", False)
        if compress is not False:
            # compress this file
            main_file = file
            file = fctutils.Compressor()

        for pair in zip(self.paramnames, self.params):
            print("%s=%.17f" % pair, file=file)

        print("maxiter=%d" % self.maxiter, file=file)
        print("yflip=%s" % self.yflip, file=file)
        print("periodicity=%s" % int(self.periodicity), file=file)
        print("period_tolerance=%.17f" % self.period_tolerance, file=file)

        self.forms[0].save_formula_params(file, self.warp_param)
        self.forms[1].save_formula_params(file)
        self.forms[2].save_formula_params(file)

        i = 0
        for transform in self.transforms:
            transform.save_formula_params(file, None, i)
            i += 1

        print("[colors]", file=file)
        print("colorizer=1", file=file)
        print("solids=[", file=file)
        for solid in self.solids:
            print("%02x%02x%02x%02x" % solid, file=file)
        print("]", file=file)
        print("[endsection]", file=file)

        if compress:
            file.close()
            print(file.getvalue(), file=main_file)

        if update_saved_flag:
            self.saved = True

    def get_gradient(self):
        try:
            g = self.forms[0].get_named_param_value("@_gradient")
            if g == 0:
                g = self.default_gradient
        except BaseException:
            g = self.default_gradient
        return g

    def set_gradient(self, g):
        old_g = self.get_gradient()
        if old_g != g:
            self.forms[0].set_gradient(g)
            #needs_redraw = False
            # if self.forms[1].is_direct() or \
            #   self.forms[2].is_direct():
            #    needs_redraw = True

            self.changed(True)  # needs_redraw)

    def set_gradient_from_file(self, file, name):
        g = self.compiler.get_gradient(file, name)
        self.set_gradient(g)

    def parse_periodicity(self, val, f):
        try:
            self.set_periodicity(int(val))
        except ValueError:
            # might be a bool in 'True'/'False' format
            self.set_periodicity(bool(val))

    def parse_period_tolerance(self, val, f):
        self.period_tolerance = float(val)

    def set_period_tolerance(self, val):
        if val != self.period_tolerance:
            self.period_tolerance = val
            self.changed(False)

    def normalize_formulafile(self, params, formindex, formtype):
        formula = params.dict.get("formula")
        if formula:
            # we have an in-line formula, use that instead of
            # formulafile/function
            (formulafile, fname) = self.compiler.add_inline_formula(
                formula, formtype)
        else:
            formulafile = params.dict.get(
                "formulafile", self.forms[formtype].funcFile)
            fname = params.dict.get(
                "function", self.forms[formtype].funcName)

        return (formulafile, fname)

    def parse__inner_(self, val, f):
        params = fctutils.ParamBag()
        params.load(f)
        (file, func) = self.normalize_formulafile(
            params, 2, fc.FormulaTypes.COLORFUNC)
        self.set_inner(file, func)
        self.forms[2].load_param_bag(params)

    def parse__outer_(self, val, f):
        params = fctutils.ParamBag()
        params.load(f)
        (file, func) = self.normalize_formulafile(
            params, 1, fc.FormulaTypes.COLORFUNC)
        self.set_outer(file, func)
        self.forms[1].load_param_bag(params)

    def parse__function_(self, val, f):
        params = fctutils.ParamBag()
        params.load(f)
        (file, func) = self.normalize_formulafile(
            params, 0, fc.FormulaTypes.FRACTAL)

        self.set_formula(file, func, 0)

        for (name, val) in list(params.dict.items()):
            if name == "formulafile" or name == "function" or name == "formula" or name == "":
                continue
            elif name == "a" or name == "b" or name == "c":
                # back-compat for older versions
                self.forms[0].set_named_param("@" + name, val)
            else:
                self.forms[0].set_named_item(name, val)

    def parse__transform_(self, val, f):
        which_transform = int(val)
        params = fctutils.ParamBag()
        params.load(f)
        self.set_transform(
            params.dict["formulafile"],
            params.dict["function"],
            which_transform)
        self.transforms[which_transform].load_param_bag(params)

    def __del__(self):
        # print "deleting fractal %s" % self
        del self.pfunc
        del self.handle

    def __copy__(self):
        # override shallow-copy to do a deeper copy than normal,
        # but still don't try and copy *everything*

        c = T(self.compiler, self.site)

        c.maxiter = self.maxiter
        c.params = copy.copy(self.params)

        c.bailfunc = self.bailfunc

        for i in range(3):
            c.set_formula(self.forms[i].funcFile, self.forms[i].funcName, i)
            c.forms[i].copy_from(self.forms[i])

        for t in self.transforms:
            c.append_transform(t.funcFile, t.funcName)
            c.transforms[-1].copy_from(t)

        c.solids = copy.copy(self.solids)
        c.yflip = self.yflip
        c.periodicity = self.periodicity
        c.period_tolerance = self.period_tolerance
        c.auto_deepen = self.auto_deepen
        c.auto_tolerance = self.auto_tolerance
        c.saved = self.saved
        c.clear_image = self.clear_image
        c.warp_param = self.warp_param
        return c

    def determine_direction(self, a, b, mode):
        isClockwise = False

        if mode == BLEND_NEAREST:
            if abs(b - a) <= math.pi and a < b:
                isClockwise = True
            elif abs(b - a) > math.pi and a > b:
                isClockwise = True
        elif mode == BLEND_FURTHEST:
            if abs(b - a) <= math.pi and a > b:
                isClockwise = True
            if abs(b - a) > math.pi and a < b:
                isClockwise = True
        elif mode == BLEND_CW:
            isClockwise = True
        elif mode == BLEND_CCW:
            isClockwise = False
        else:
            raise ValueError("Unknown angle blend mode %s" % mode)

        return isClockwise

    def blend_angle(self, a, b, ratio, mode):
        isClockwise = self.determine_direction(a, b, mode)

        if isClockwise and b < a:
            b = b + math.pi * 2.0
        if not isClockwise and b > a:
            a = a + math.pi * 2.0

        angle = a * (1 - ratio) + b * ratio
        while angle > math.pi:
            angle -= math.pi * 2.0

        return angle

    def blend(self, other, ratio, angle_options=()):
        """Create a new fractal which blends this and the other's parameter sets using ratio.
        'angle_options' can be used to override the default method of interpolating angles."""
        new = copy.copy(self)
        for i in range(self.XCENTER, self.MAGNITUDE):
            a, b = self.params[i], other.params[i]
            new.set_param(i, a * (1 - ratio) + b * ratio)

        # magnitude is exponential
        a, b = self.params[self.MAGNITUDE], other.params[self.MAGNITUDE]
        if abs(a) > abs(b):
            factor = (a - b) / (math.e - 1)
            val = factor * math.exp(1 - ratio) + (b - factor)
        else:
            factor = (b - a) / (math.e - 1)
            val = factor * math.exp(ratio) + (a - factor)

        new.set_param(self.MAGNITUDE, val)

        for i in range(self.XYANGLE, self.ZWANGLE + 1):
            a, b = self.params[i], other.params[i]
            option = angle_options[i - self.XYANGLE:i - self.XYANGLE]
            if option:
                mode = option[0]
            else:
                mode = BLEND_NEAREST
            new.set_param(i, self.blend_angle(a, b, ratio, mode))

        for form1, form2 in zip(new.forms, other.forms):
            form1.blend(form2, ratio)

        return new

    def reset_angles(self):
        for i in range(self.XYANGLE, self.ZWANGLE + 1):
            self.set_param(i, 0.0)

    def reset(self):
        # set global default values, then override from formula
        # set up defaults
        self.params = [
            0.0, 0.0, 0.0, 0.0,  # center
            4.0,  # size
            0.0, 0.0, 0.0, 0.0, 0.0, 0.0  # angles
        ]

        self.bailout = 0.0
        self.maxiter = 256
        self.rot_by = math.pi / 2
        self.title = self.forms[0].funcName
        self.yflip = False
        self.auto_epsilon = False
        self.period_tolerance = 1.0E-9

        g = self.get_gradient()
        self.set_formula_defaults(g)

    def reset_zoom(self):
        mag = self.forms[0].formula.defaults.get("magn")
        if mag:
            mag = mag.value
        else:
            mag = self.forms[0].formula.defaults.get("magnitude")
            if mag:
                mag = mag.value
            else:
                mag = 4.0
        self.set_param(self.MAGNITUDE, mag)

    def copy_colors(self, f):
        self.set_gradient(copy.copy(f.get_gradient()))
        self.set_solids(f.solids)
        self.changed(False)

    def set_warp_param(self, param):
        if self.warp_param != param:
            self.warp_param = param
            self.changed(True)

    def set_cmap(self, mapfile):
        c = colorizer.T(self)
        with open(mapfile) as file:
            c.parse_map_file(file)
        self.set_gradient(c.gradient)
        self.set_solids(c.solids)
        self.changed(False)

    def get_initparam(self, n, param_type):
        params = self.forms[param_type].params
        return params[n]

    def set_initparam(self, n, val, param_type):
        self.forms[param_type].set_param(n, val)

    def set_solid(self, i, newsolid):
        if self.solids[i] == newsolid:
            return
        self.solids[i] = newsolid
        self.changed(False)

    def set_solids(self, solids):
        same = True
        for i in range(len(solids)):
            if self.solids[i] != solids[i]:
                same = False
                break
        if same:
            return

        self.solids[0:len(solids)] = solids[:]
        self.changed(False)

    def refresh(self):
        for i in range(3):
            if self.compiler.out_of_date(self.forms[i].funcFile):
                self.set_formula(
                    self.forms[i].funcFile, self.forms[i].funcName, i)

    def set_formula_defaults(self, g=None):
        if self.forms[0].formula is None:
            return

        if g is None:
            g = self.get_gradient()

        self.forms[0].set_initparams_from_formula(g)

        for (name, val) in list(self.forms[0].formula.defaults.items()):
            # FIXME helpfile,helptopic,method,precision,
            # render,skew,stretch
            if name == "maxiter":
                self.maxiter = int(val.value)
            elif name == "center" or name == "xycenter":
                self.params[self.XCENTER] = float(val.value[0].value)
                self.params[self.YCENTER] = float(val.value[1].value)
            elif name == "zwcenter":
                self.params[self.ZCENTER] = float(val.value[0].value)
                self.params[self.WCENTER] = float(val.value[1].value)
            elif name == "angle":
                self.params[self.XYANGLE] = float(val.value)
            elif name == "magn":
                self.params[self.MAGNITUDE] = float(val.value)
            elif name == "title":
                self.title = val.value
            elif name == "periodicity":
                self.periodicity = int(val.value)
            else:
                if hasattr(self, name.upper()):
                    self.params[getattr(self, name.upper())] = float(val.value)
                else:
                    print("ignored unknown parameter %s" % name)

        for form in self.forms[1:] + self.transforms:
            form.reset_params()

    def set_formula(self, formulafile, func, index=0):
        self.forms[index].set_formula(formulafile, func, self.get_gradient())

        if index == 0:
            self.set_bailfunc()
            self.warp_param = None
        self.formula_changed()
        self.changed()

    def get_saved(self):
        return self.saved

    def set_bailfunc(self):
        bailfuncs = [
            "cmag", "manhattanish", "manhattanish2",
            "max2", "min2",
            "real2", "imag2",
            None  # bailout
        ]
        funcname = bailfuncs[self.bailfunc]
        if funcname is None:
            # FIXME deal with diff
            return

        func = self.forms[0].formula.symbols.get("@bailfunc")
        if func is not None:
            self.set_func(func[0], funcname, self.forms[0].formula)

    def changed(self, clear_image=True):
        self.dirty = True
        self.saved = False
        self.clear_image = clear_image

    def formula_changed(self):
        self.dirtyFormula = True

    def set_func(self, func, fname, formula):
        if func.cname != fname:
            formula.symbols.set_std_func(func, fname)
            self.dirtyFormula = True
            self.changed()

    def set_periodicity(self, periodicity):
        if self.periodicity != periodicity:
            self.periodicity = periodicity
            self.changed()

    def set_inner(self, funcfile, funcname):
        self.set_formula(funcfile, funcname, 2)

    def set_outer(self, funcfile, funcname):
        self.set_formula(funcfile, funcname, 1)

    def get_transform_prefix(self):
        i = self.next_transform_id
        prefix = "t%d" % i
        self.next_transform_id += 1
        return prefix

    def append_transform(self, funcfile, funcname):
        fs = formsettings.T(self.compiler, self, self.get_transform_prefix())
        fs.set_formula(funcfile, funcname, self.get_gradient())
        self.transforms.append(fs)
        self.formula_changed()
        self.changed()

    def set_transform(self, funcfile, funcname, i):
        fs = formsettings.T(self.compiler, self, self.get_transform_prefix())
        fs.set_formula(funcfile, funcname, self.get_gradient())
        if len(self.transforms) <= i:
            self.transforms.extend([None] * (i - len(self.transforms) + 1))

        self.transforms[i] = fs
        self.formula_changed()
        self.changed()

    def remove_transform(self, i):
        self.transforms.pop(i)
        self.formula_changed()
        self.changed()

    def set_compiler_option(self, option, val):
        self.compiler_options[option] = val
        self.dirtyFormula = True

    def apply_options(self, options):
        if options.formula.name and options.formula.func:
            self.set_formula(options.formula.name, options.formula.func)
            self.reset()

        if options.inner.name and options.inner.func:
            self.set_formula(options.inner.name, options.inner.func)
            self.reset()

        if options.outer.name and options.outer.func:
            self.set_formula(options.outer.name, options.outer.func)
            self.reset()

        if options.maxiter != -1:
            self.set_maxiter(options.maxiter)

        for num, val in options.paramchanges.items():
            self.set_param(num, val)

        for t in options.transforms:
            self.append_transform(t.name, t.func)

        if options.map:
            self.set_cmap(options.map)

        if options.antialias is not None:
            self.antialias = options.antialias

    def compile(self):
        if self.forms[0].formula is None:
            raise ValueError("no formula")
        if self.dirtyFormula is False:
            return self.outputfile

        outputfile = self.compiler.compile_all(
            self.forms[0].formula,
            self.forms[1].formula,
            self.forms[2].formula,
            [x.formula for x in self.transforms],
            self.compiler_options)

        if outputfile is not None:
            self.set_output_file(outputfile)

        self.dirtyFormula = False
        return self.outputfile

    def set_output_file(self, outputfile):
        if self.outputfile != outputfile:
            self.outputfile = outputfile
            self.handle = fract4dc.pf_load(self.outputfile)
            self.pfunc = fract4dc.pf_create(self.handle)

    def make_random_colors(self, n):
        self.get_gradient().randomize(n)
        self.changed(False)

    def mul_vs(self, v, s):
        return [x * s for x in v]

    def xy_random(self, weirdness, size):
        return weirdness * 0.5 * size * (random.random() - 0.5)

    def zw_random(self, weirdness, size):
        factor = math.fabs(1.0 - math.log(size)) + 1.0
        return weirdness * (random.random() - 0.5) * 1.0 / factor

    def angle_random(self, weirdness):
        action = random.random()
        if action > weirdness:
            return 0.0  # no change

        action = random.random()
        if action < weirdness / 6.0:
            # +/- pi/2
            if random.random() > 0.5:
                return math.pi / 2.0
            else:
                return math.pi / 2.0

        return weirdness * (random.random() - 0.5) * math.pi / 2.0

    def is4D(self):
        return self.warp_param is not None or self.forms[0].formula.is4D()

    def mutate(self, weirdness, color_weirdness):
        '''randomly adjust position, colors, angles and parameters.
        weirdness is between 0 and 1 - 0 is no change, 1 is lots'''

        size = self.params[self.MAGNITUDE]
        self.params[self.XCENTER] += self.xy_random(weirdness, size)
        self.params[self.YCENTER] += self.xy_random(weirdness, size)

        self.params[self.XYANGLE] += self.angle_random(weirdness)

        if self.is4D():
            self.params[self.ZCENTER] += self.zw_random(weirdness, size)
            self.params[self.WCENTER] += self.zw_random(weirdness, size)

            for a in range(self.XZANGLE, self.ZWANGLE):
                self.params[a] += self.angle_random(weirdness)

        if random.random() < weirdness * 0.75:
            self.params[self.MAGNITUDE] *= 1.0 + (0.5 - random.random())

        for f in self.forms:
            f.mutate(weirdness, size)

        if random.random() < color_weirdness:
            try:
                (file, formula) = self.compiler.get_random_gradient()
                self.set_gradient_from_file(file, formula)
            except IndexError:
                # can occur if no gradients available or occasionally
                # because random.choice is horked
                pass

    def nudge(self, x, y, axis=0):
        # move a little way in x or y
        self.relocate(0.025 * x, 0.025 * y, 1.0, axis)

    def get_form(self, param_type):
        if param_type > 2:
            form = self.transforms[param_type - 3]
        else:
            form = self.forms[param_type]
        return form

    def nudge_param(self, i, param_type, x, y):
        form = self.get_form(param_type)
        form.nudge_param(i, x, y)

    def relocate(self, dx, dy, zoom, axis=0):
        if dx == 0 and dy == 0 and zoom == 1.0:
            return

        m = fract4dc.rot_matrix(self.params)

        deltax = self.mul_vs(m[axis], dx)
        if self.yflip:
            deltay = self.mul_vs(m[axis + 1], dy)
        else:
            deltay = self.mul_vs(m[axis + 1], -dy)

        self.params[self.XCENTER] += deltax[0] + deltay[0]
        self.params[self.YCENTER] += deltax[1] + deltay[1]
        self.params[self.ZCENTER] += deltax[2] + deltay[2]
        self.params[self.WCENTER] += deltax[3] + deltay[3]
        self.params[self.MAGNITUDE] *= zoom
        self.changed()

    def flip_to_julia(self):
        self.params[self.XZANGLE] += self.rot_by
        self.params[self.YWANGLE] += self.rot_by
        self.rot_by = - self.rot_by
        self.changed()

    # status callbacks
    def status_changed(self, val):
        pass

    def progress_changed(self, d):
        pass

    def stats_changed(self, s):
        pass

    def is_interrupted(self):
        return False

    def iters_changed(self, iters):
        # print "iters changed to %d" % iters
        self.maxiter = iters

    def tolerance_changed(self, tolerance):
        # print "tolerance changed to %g" % tolerance
        self.period_tolerance = tolerance

    def image_changed(self, x1, y1, x2, y2):
        pass

    # you need to rename this to pixel_changed to get pixel debug info
    def _pixel_changed(self, params, x, y, aa, maxIters,
                       nNoPeriodIters, dist, fate, nIters, r, g, b, a):
        # remove underscore to debug fractal generation
        print("pixel: (%g,%g,%g,%g) %d %d %d %d %d %g %d %d (%d %d %d %d)" %
              (params[0], params[1], params[2], params[3], x, y, aa, maxIters, nNoPeriodIters, dist, fate, nIters, r, g, b, a))

    def epsilon_tolerance(self, w, h):
        # 5% of the size of a pixel
        return self.params[self.MAGNITUDE] / (20.0 * max(w, h))

    def all_params(self):
        p = []
        for form in self.forms:
            p += form.params
        for transform in self.transforms:
            p += transform.params
        return p

    def get_colormap(self):
        cmap = fract4dc.cmap_create_gradient(self.get_gradient().segments)

        (r, g, b, a) = self.solids[0]
        fract4dc.cmap_set_solid(cmap, 0, r, g, b, a)
        (r, g, b, a) = self.solids[1]
        fract4dc.cmap_set_solid(cmap, 1, r, g, b, a)

        return cmap

    def init_pfunc(self):
        initparams = self.all_params()
        fract4dc.pf_init(self.pfunc, self.params, initparams)

    def get_warp(self):
        if self.warp_param:
            warp = self.forms[0].order_of_name(self.warp_param)
        else:
            warp = -1
        return warp

    def set_maxiter(self, new_iter):
        if self.maxiter != new_iter:
            self.maxiter = new_iter
            self.changed(False)

    def set_antialias(self, aa):
        if aa != self.antialias:
            self.antialias = aa
            self.changed(True)

    def set_auto_deepen(self, auto_deepen):
        if auto_deepen != self.auto_deepen:
            self.auto_deepen = auto_deepen
            self.changed(True)

    def set_auto_tolerance(self, auto_tolerance):
        if auto_tolerance != self.auto_tolerance:
            self.auto_tolerance = auto_tolerance
            self.changed(True)

    def calc(self, image, colormap, nthreads, site, asynchronous):
        fract4dc.calc(
            params=self.params,
            antialias=self.antialias,
            maxiter=self.maxiter,
            yflip=self.yflip,
            periodicity=self.periodicity,
            nthreads=nthreads,
            pfo=self.pfunc,
            cmap=colormap,
            auto_deepen=self.auto_deepen,
            auto_tolerance=self.auto_tolerance,
            tolerance=self.period_tolerance,
            render_type=self.render_type,
            warp_param=self.get_warp(),
            image=image._img,
            site=site,
            dirty=self.clear_image,
            asynchronous=asynchronous)

    def drawpoint(self, repeats=1000):
        self.init_pfunc()
        print("x:\t\t%.17f\ny:\t\t%.17f\nz:\t\t%.17f\nw:\t\t%.17f\n" %
              tuple(self.params[0:4]))
        startTime = now()
        result = fract4dc.pf_calc(
            self.pfunc, self.params[0:4], self.maxiter, 0, 0, 0, repeats)
        duration = now() - startTime
        print(
            "iterations:\t%s\nfate:\t\t%s\ndistance:\t%s\nsolid:\t\t%s" %
            result)
        print("duration:\t%.4g" % duration)

    def draw(self, image, nthreads=1):
        self.init_pfunc()

        colormap = self.get_colormap()
        for (xoff, yoff, xres, yres) in image.get_tile_list():
            image.resize_tile(xres, yres)
            image.set_offset(xoff, yoff)

            self.calc(image, colormap, nthreads, self.site, False)

            image.save_tile()

    def clean(self):
        self.dirty = False

    def set_param(self, n, val):
        val = float(val)
        if self.params[n] != val:
            self.params[n] = val
            self.changed()

    def get_param(self, n):
        return self.params[n]

    def parse_gnofract4d_parameter_file(self, val, f):
        pass

    def parse_version_string(self, s):
        try:
            (major, minor) = tuple([int(a) for a in s.split(".")])
            return major * 1000.0 + minor
        except BaseException:
            raise ValueError("Invalid version number %s" % s)

    def parse_version(self, val, f):
        global THIS_FORMAT_VERSION
        self.format_version = self.parse_version_string(val)
        this_format_version = self.parse_version_string(THIS_FORMAT_VERSION)
        if self.format_version < 2000.0:
            # old versions displayed everything upside down
            # switch the rotation so they load OK
            self.yflip = True
        if 1700.0 < self.format_version < 2000.0:
            # a version that used auto-tolerance for Nova and Newton
            self.auto_epsilon = True

        if self.format_version > this_format_version:
            warning = \
                '''This file was created by a newer version of Gnofract 4D.
The image may not display correctly. Please upgrade to version %s or higher.'''

            self.warn(warning % val)

    def warn(self, msg):
        print(msg)

    def parse_bailfunc(self, val, f):
        # can't set function directly because formula hasn't been parsed yet
        self.bailfunc = int(val)

    def apply_colorizer(self, cf):
        if cf.read_gradient:
            self.set_gradient(cf.gradient)
        self.set_solids(cf.solids)
        self.changed(False)
        if cf.direct:
            # loading a legacy rgb colorizer
            self.set_outer("gf4d.cfrm", "rgb")

            val = "(%f,%f,%f,1.0)" % tuple(cf.rgb)
            self.forms[1].set_named_item("@col", val)

    def parse__colors_(self, val, f):
        cf = colorizer.T(self)
        cf.load(f)
        self.apply_colorizer(cf)

    def parse__colorizer_(self, val, f):
        which_cf = int(val)
        cf = colorizer.T(self)
        cf.load(f)
        if which_cf == 0:
            self.apply_colorizer(cf)
        # ignore other colorlists for now

    def parse_inner(self, val, f):
        name = self.colorfunc_names[int(val)]
        self.set_inner("gf4d.cfrm", name)

    def parse_outer(self, val, f):
        name = self.colorfunc_names[int(val)]
        self.set_outer("gf4d.cfrm", name)

    def set_yflip(self, yflip):
        if self.yflip != yflip:
            self.yflip = yflip
            self.changed()

    def parse_yflip(self, val, f):
        self.yflip = (val == "1" or val == "True")

    def parse_x(self, val, f):
        self.set_param(self.XCENTER, val)

    def parse_y(self, val, f):
        self.set_param(self.YCENTER, val)

    def parse_z(self, val, f):
        self.set_param(self.ZCENTER, val)

    def parse_w(self, val, f):
        self.set_param(self.WCENTER, val)

    def parse_size(self, val, f):
        self.set_param(self.MAGNITUDE, val)

    def parse_xy(self, val, f):
        self.set_param(self.XYANGLE, val)

    def parse_xz(self, val, f):
        self.set_param(self.XZANGLE, val)

    def parse_xw(self, val, f):
        self.set_param(self.XWANGLE, val)

    def parse_yz(self, val, f):
        self.set_param(self.YZANGLE, val)

    def parse_yw(self, val, f):
        self.set_param(self.YWANGLE, val)

    def parse_zw(self, val, f):
        self.set_param(self.ZWANGLE, val)

    def parse_bailout(self, val, f):
        self.bailout = float(val)

    def parse_maxiter(self, val, f):
        self.maxiter = int(val)

    def parse_antialias(self, val, f):
        # antialias now a user pref, not saved in file
        #self.antialias = int(val)
        pass

    def order_of_name(self, name, symbol_table):
        op = symbol_table.order_of_params()
        rn = symbol_table.mangled_name(name)
        ord = op.get(rn)
        if ord is None:
            # print "can't find %s (%s) in %s" % (name,rn,op)
            pass
        return ord

    def fix_bailout(self):
        # because bailout occurs before we know which function this is
        # in older files, we save it in self.bailout then apply to the
        # initparams later
        if self.bailout != 0.0:
            for f in self.forms:
                f.try_set_named_item("@bailout", self.bailout)

    def fix_gradients(self, old_gradient):
        # new gradient is read in after the gradient params have been set,
        # so this is needed to fix any which are using that default
        p = self.forms[0].params
        for i in range(len(p)):
            if p[i] == old_gradient:
                p[i] = self.get_gradient()

    def param_display_name(self, name, param):
        if hasattr(param, "title"):
            return param.title.value
        if hasattr(param, "caption"):
            return param.caption.value
        if name[:5] == "t__a_":
            name = name[5:]
        return name.title()

    def param_tip(self, name, param):
        if hasattr(param, "hint"):
            return param.hint.value
        return self.param_display_name(name, param)

    def loadFctFile(self, f):
        #old_gradient = self.get_gradient()
        line = f.readline()
        if line is None or not line.startswith("gnofract4d parameter file"):
            raise Exception("Not a valid parameter file")

        self.load(f)

        self.fix_bailout()
        # self.fix_gradients(old_gradient)
        self.saved = True



# a grab-bag of simple optimization techniques

# while we rely on the C compiler to do the heavy lifting,
# it's reluctant to do anything to floating point arithmetic
# in case it changes the behavior of the program. We have
# a freer hand because
# a) this is for drawing pretty pictures, not calculating missile
#    trajectories or anything
# b) we always ignore overflow, NaN, etc

from . import instructions, graph

Nothing = 0
Peephole = 1
ConstantPropagation = 2


class FlowGraph:
    "Builds a control flow graph from a sequence of instructions"

    def __init__(self):
        pass

    def build(self, insns):
        self.control = graph.T()
        self.define = {}
        self.use = {}
        for insn in insns:
            if isinstance(insn, instructions.Insn):
                n = self.control.newNode()
                for d in insn.dest():
                    self.define.setdefault(n, []).append(d)
                for s in insn.source():
                    self.use.setdefault(n, []).append(s)


class T:
    "Holds overall optimization logic"

    def __init__(self):
        pass

    def peephole_binop(self, insn):
        left = insn.src[0]
        right = insn.src[1]
        if isinstance(left, instructions.ConstArg):
            if isinstance(right, instructions.ConstArg):
                # both args constant, replace with new const
                return insn.const_eval()
            else:
                const_index = 0
                other_index = 1
        else:
            if isinstance(right, instructions.ConstArg):
                const_index = 1
                other_index = 0
            else:
                # neither are constant, we can't do anything
                return insn

        if insn.op == "*":
            # 1 * n => n
            if insn.src[const_index].is_one():
                return instructions.Move(
                    [insn.src[other_index]],
                    insn.dst)

            # 0 * n -> 0
            if insn.src[const_index].is_zero():
                return instructions.Move(
                    [insn.src[const_index]],
                    insn.dst)

        return insn

    def peephole_insn(self, insn):
        """Return an optimized version of this instruction, if possible,
        or the same instruction if no change is practical."""
        if isinstance(insn, instructions.Binop):
            return self.peephole_binop(insn)
        else:
            return insn

    def peephole(self, insns):
        """Perform some straightforward algebraic simplifications
        which don't require any global knowledge."""

        out_insns = []

        for insn in insns:
            result = self.peephole_insn(insn)
            if result:
                out_insns.append(result)

        return out_insns

    def constant_propagation(self, insns):
        """if we have:

        t = const
        a = t op b

        Then we can replace that with

        a = const op b

        This is done using dataflow analysis."""

    def optimize(self, flags, insns):
        if (flags & Peephole):
            insns = self.peephole(insns)

        return insns



# The fractal standard library, including operators
import math

from .instructions import ComplexArg, ConstFloatArg, ConstIntArg, TempArg, HyperArg, ColorArg
from .fracttypes import *


class Constants:
    i = ComplexArg(ConstFloatArg(0.0), ConstFloatArg(1.0))
    iby2 = ComplexArg(ConstFloatArg(0.0), ConstFloatArg(0.5))
    minus_i = ComplexArg(ConstFloatArg(0.0), ConstFloatArg(-1.0))
    one = ComplexArg(ConstFloatArg(1.0), ConstFloatArg(0.0))


def reals(l):
    # [[a + ib], [c+id]] => [ a, c]
    return [x.re for x in l]


def imags(l):
    return [x.im for x in l]


def parts(n, l):
    return [x.parts[n] for x in l]


def not_b_b(gen, t, srcs):
    return gen.emit_func('!', srcs, Int)

# unary negation


def neg_i_i(gen, t, srcs):
    return gen.emit_func('-', srcs, Int)


def neg_f_f(gen, t, srcs):
    return gen.emit_func('-', srcs, Float)


def neg_c_c(gen, t, srcs):
    return ComplexArg(
        gen.emit_func('-', [srcs[0].re], Float),
        gen.emit_func('-', [srcs[0].im], Float))


def neg_h_h(gen, t, srcs):
    src = srcs[0]
    return HyperArg(
        gen.emit_func('-', [src.parts[0]], Float),
        gen.emit_func('-', [src.parts[1]], Float),
        gen.emit_func('-', [src.parts[2]], Float),
        gen.emit_func('-', [src.parts[3]], Float))

# basic binary operation


def add_ff_f(gen, t, srcs):
    return gen.emit_binop(t.op, srcs, t.datatype)


# many equivalent funcs
sub_ff_f = mul_ff_f = div_ff_f = add_ff_f
add_ii_i = sub_ii_i = mul_ii_i = div_ii_i = add_ff_f
gt_ii_b = gte_ii_b = lt_ii_b = lte_ii_b = eq_ii_b = noteq_ii_b = add_ff_f
gt_ff_b = gte_ff_b = lt_ff_b = lte_ff_b = eq_ff_b = noteq_ff_b = add_ff_f


def mod_ii_i(gen, t, srcs):
    return gen.emit_binop('%', srcs, t.datatype)


def mod_ff_f(gen, t, srcs):
    return gen.emit_func2('fmod', srcs, Float)


def mul_cc_c(gen, t, srcs):
    # (a+ib) * (c+id) = ac - bd + i(bc + ad)
    ac = gen.emit_binop('*', [srcs[0].re, srcs[1].re], Float)
    bd = gen.emit_binop('*', [srcs[0].im, srcs[1].im], Float)
    bc = gen.emit_binop('*', [srcs[0].im, srcs[1].re], Float)
    ad = gen.emit_binop('*', [srcs[0].re, srcs[1].im], Float)
    dst = ComplexArg(
        gen.emit_binop('-', [ac, bd], Float),
        gen.emit_binop('+', [bc, ad], Float))
    return dst


def mul_hh_h(gen, t, srcs):
    def mulpair(x, y):
        return gen.emit_binop('*', [srcs[0].parts[x], srcs[1].parts[y]], Float)

    re2 = mulpair(0, 0)
    i2 = mulpair(1, 1)
    j2 = mulpair(2, 2)
    k2 = mulpair(3, 3)

    re = gen.emit_binop('-', [re2, i2], Float)
    re = gen.emit_binop('-', [re, j2], Float)
    re = gen.emit_binop('+', [re, k2], Float)

    i_re = mulpair(1, 0)
    re_i = mulpair(0, 1)
    k_j = mulpair(3, 2)
    j_k = mulpair(2, 3)

    i = gen.emit_binop('+', [i_re, re_i], Float)
    i = gen.emit_binop('-', [i, k_j], Float)
    i = gen.emit_binop('-', [i, j_k], Float)

    j_re = mulpair(2, 0)
    k_i = mulpair(3, 1)
    re_j = mulpair(0, 2)
    i_k = mulpair(1, 3)

    j = gen.emit_binop('-', [j_re, k_i], Float)
    j = gen.emit_binop('+', [j, re_j], Float)
    j = gen.emit_binop('-', [j, i_k], Float)

    k_re = mulpair(3, 0)
    j_i = mulpair(2, 1)
    i_j = mulpair(1, 2)
    re_k = mulpair(0, 3)

    k = gen.emit_binop('+', [k_re, j_i], Float)
    k = gen.emit_binop('+', [k, i_j], Float)
    k = gen.emit_binop('+', [k, re_k], Float)

    return HyperArg(re, i, j, k)


def complex_ff_c(gen, t, srcs):
    # construct a complex number from two real parts
    return ComplexArg(srcs[0], srcs[1])


def hyper_ffff_h(gen, t, srcs):
    return HyperArg(srcs[0], srcs[1], srcs[2], srcs[3])


def hyper_cc_h(gen, t, srcs):
    return HyperArg(srcs[0].re, srcs[0].im, srcs[1].re, srcs[1].im)


def add_cc_c(gen, t, srcs):
    # add 2 complex numbers
    dst = ComplexArg(
        gen.emit_binop('+', reals(srcs), Float),
        gen.emit_binop('+', imags(srcs), Float))
    return dst


def add_hh_h(gen, t, srcs):
    # add 2 hyper numbers
    dst = HyperArg(
        gen.emit_binop('+', parts(0, srcs), Float),
        gen.emit_binop('+', parts(1, srcs), Float),
        gen.emit_binop('+', parts(2, srcs), Float),
        gen.emit_binop('+', parts(3, srcs), Float))
    return dst


def clamp_f_f(gen, t, srcs):
    # ensure f is in the range [0,1.0]
    smaller_than_one = gen.symbols.newLabel()
    done = gen.symbols.newLabel()

    one = ConstFloatArg(1.0)
    zero = ConstFloatArg(0.0)
    src = srcs[0]
    dst = TempArg(gen.symbols.newTemp(Float), Float)

    gen.emit_move(src, dst)

    lte1 = gen.emit_binop('<=', [src, one], Float)
    gen.emit_cjump(lte1, smaller_than_one)

    # larger than 1: use one instead
    gen.emit_move(one, dst)
    gen.emit_jump(done)

    gen.emit_label(smaller_than_one)
    gte0 = gen.emit_binop('>=', [src, zero], Float)
    gen.emit_cjump(gte0, done)

    # smaller than 0: use 0 instead
    gen.emit_move(zero, dst)

    gen.emit_label(done)

    return dst


def clamp_C_C(gen, t, srcs):
    # ensure color C is in the range [0.0,1.0]
    dst = ColorArg(
        clamp_f_f(gen, t, parts(0, srcs)),
        clamp_f_f(gen, t, parts(1, srcs)),
        clamp_f_f(gen, t, parts(2, srcs)),
        clamp_f_f(gen, t, parts(3, srcs)))
    return dst


def add_CC_C(gen, t, srcs):
    # add 2 colors, piecewise
    dst = ColorArg(
        gen.emit_binop('+', parts(0, srcs), Float),
        gen.emit_binop('+', parts(1, srcs), Float),
        gen.emit_binop('+', parts(2, srcs), Float),
        gen.emit_binop('+', parts(3, srcs), Float))
    return dst


def sub_cc_c(gen, t, srcs):
    # subtract 2 complex numbers
    dst = ComplexArg(
        gen.emit_binop('-', reals(srcs), Float),
        gen.emit_binop('-', imags(srcs), Float))
    return dst


def sub_hh_h(gen, t, srcs):
    # subtract 2 hypercomplex numbers
    dst = HyperArg(
        gen.emit_binop('-', parts(0, srcs), Float),
        gen.emit_binop('-', parts(1, srcs), Float),
        gen.emit_binop('-', parts(2, srcs), Float),
        gen.emit_binop('-', parts(3, srcs), Float))
    return dst


def sub_CC_C(gen, t, srcs):
    dst = ColorArg(
        gen.emit_binop('-', parts(0, srcs), Float),
        gen.emit_binop('-', parts(1, srcs), Float),
        gen.emit_binop('-', parts(2, srcs), Float),
        gen.emit_binop('-', parts(3, srcs), Float))
    return dst


def div_cc_c(gen, t, srcs):
    # (a+ib)/(c+id) = (a+ib)*(c-id) / (c+id)*(c-id)
    # = (ac + bd + i(bc - ad))/mag(c+id)
    denom = cmag_c_f(gen, 'mag', [srcs[1]])
    ac = gen.emit_binop('*', [srcs[0].re, srcs[1].re], Float)
    bd = gen.emit_binop('*', [srcs[0].im, srcs[1].im], Float)
    bc = gen.emit_binop('*', [srcs[0].im, srcs[1].re], Float)
    ad = gen.emit_binop('*', [srcs[0].re, srcs[1].im], Float)
    dre = gen.emit_binop('+', [ac, bd], Float)
    dim = gen.emit_binop('-', [bc, ad], Float)
    return ComplexArg(
        gen.emit_binop('/', [dre, denom], Float),
        gen.emit_binop('/', [dim, denom], Float))


def div_cf_c(gen, t, srcs):
    # divide a complex number by a real one
    return ComplexArg(
        gen.emit_binop('/', [srcs[0].re, srcs[1]], Float),
        gen.emit_binop('/', [srcs[0].im, srcs[1]], Float))


def mul_cf_c(gen, t, srcs):
    # multiply a complex number by a real one
    return ComplexArg(
        gen.emit_binop('*', [srcs[0].re, srcs[1]], Float),
        gen.emit_binop('*', [srcs[0].im, srcs[1]], Float))


def mul_hf_h(gen, t, srcs):
    # multiply a hypercomplex number by a real one
    return HyperArg(
        gen.emit_binop('*', [srcs[0].parts[0], srcs[1]], Float),
        gen.emit_binop('*', [srcs[0].parts[1], srcs[1]], Float),
        gen.emit_binop('*', [srcs[0].parts[2], srcs[1]], Float),
        gen.emit_binop('*', [srcs[0].parts[3], srcs[1]], Float)
    )


def mul_Cf_C(gen, t, srcs):
    return mul_hf_h(gen, t, srcs)


def div_hf_h(gen, t, srcs):
    # multiply a hypercomplex number by a real one
    return HyperArg(
        gen.emit_binop('/', [srcs[0].parts[0], srcs[1]], Float),
        gen.emit_binop('/', [srcs[0].parts[1], srcs[1]], Float),
        gen.emit_binop('/', [srcs[0].parts[2], srcs[1]], Float),
        gen.emit_binop('/', [srcs[0].parts[3], srcs[1]], Float)
    )


def div_Cf_C(gen, t, srcs):
    return div_hf_h(gen, t, srcs)


def cmag_c_f(gen, t, srcs):
    # |x| = x_re * x_re + x_im * x_im
    src = srcs[0]
    re_2 = gen.emit_binop('*', [src.re, src.re], Float)
    im_2 = gen.emit_binop('*', [src.im, src.im], Float)
    return gen.emit_binop('+', [re_2, im_2], Float)


def cmag_h_f(gen, t, srcs):
    # |x| = x_re * x_re + x_i * x_i + x_j * x_j + x_k * x_k
    src = srcs[0]
    re_2 = gen.emit_binop('*', [src.parts[0], src.parts[0]], Float)
    i_2 = gen.emit_binop('*', [src.parts[1], src.parts[1]], Float)
    j_2 = gen.emit_binop('*', [src.parts[2], src.parts[2]], Float)
    k_2 = gen.emit_binop('*', [src.parts[3], src.parts[3]], Float)
    ret = gen.emit_binop('+', [re_2, i_2], Float)
    ret = gen.emit_binop('+', [ret, j_2], Float)
    ret = gen.emit_binop('+', [ret, k_2], Float)
    return ret


def log_f_f(gen, t, srcs):
    return gen.emit_func('log', srcs, Float)


def log_c_c(gen, t, srcs):
    # log(a+ib) = (log(mag(a+ib)), atan2(a+ib))
    re = gen.emit_func('log', [cabs_c_f(gen, t, srcs)], Float)
    im = atan2_c_f(gen, t, srcs)
    return ComplexArg(re, im)


def polar_ff_c(gen, t, srcs):
    # polar(r,theta) = (r * cos(theta), r * sin(theta))
    re = gen.emit_binop('*', [srcs[0], cos_f_f(gen, t, [srcs[1]])], Float)
    im = gen.emit_binop('*', [srcs[0], sin_f_f(gen, t, [srcs[1]])], Float)
    return ComplexArg(re, im)


def exp_f_f(gen, t, srcs):
    return gen.emit_func('exp', srcs, Float)


def exp_c_c(gen, t, srcs):
    # exp(a+ib) = polar(exp(a),b)
    expx = gen.emit_func('exp', [srcs[0].re], Float)
    return polar_ff_c(gen, t, [expx, srcs[0].im])


def pow_ff_f(gen, t, srcs):
    return gen.emit_func2('pow', srcs, Float)


def pow_ff_c(gen, t, srcs):
    arg = ComplexArg(srcs[0], ConstFloatArg(0.0))
    return pow_cf_c(gen, t, [arg, srcs[1]])


def pow_cf_c(gen, t, srcs):
    nonzero = gen.symbols.newLabel()
    done = gen.symbols.newLabel()
    dst_re = gen.newTemp(Float)
    dst_im = gen.newTemp(Float)

    # FIXME: this shortcut breaks 3damand01, not sure why
    # gen.emit_cjump(srcs[0].im,nonzero)

    # compute result if just real
    #tdest = pow_ff_f(gen,t,[srcs[0].re,srcs[1]])
    # gen.emit_move(tdest,dst_re)
    # gen.emit_move(ConstFloatArg(0.0),dst_im)
    # gen.emit_jump(done)

    # gen.emit_label(nonzero)\

    # result if real + imag
    # temp = log(a+ib)
    # polar(y * real(temp), y * imag(temp))

    temp = log_c_c(gen, t, [srcs[0]])
    t_re = gen.emit_binop('*', [temp.re, srcs[1]], Float)
    t_re = gen.emit_func('exp', [t_re], Float)
    t_im = gen.emit_binop('*', [temp.im, srcs[1]], Float)
    temp2 = polar_ff_c(gen, t, [t_re, t_im])
    gen.emit_move(temp2.re, dst_re)
    gen.emit_move(temp2.im, dst_im)
    gen.emit_label(done)

    return ComplexArg(dst_re, dst_im)


def pow_cc_c(gen, t, srcs):
    nonzero = gen.symbols.newLabel()
    done = gen.symbols.newLabel()
    dst_re = gen.newTemp(Float)
    dst_im = gen.newTemp(Float)

    gen.emit_cjump(srcs[0].re, nonzero)
    gen.emit_cjump(srcs[0].im, nonzero)

    # 0^foo = 0
    gen.emit_move(ConstFloatArg(0.0), dst_re)
    gen.emit_move(ConstFloatArg(0.0), dst_im)
    gen.emit_jump(done)

    gen.emit_label(nonzero)
    # exp(y*log(x))

    logx = log_c_c(gen, t, [srcs[0]])
    ylogx = mul_cc_c(gen, t, [srcs[1], logx])
    xtoy = exp_c_c(gen, t, [ylogx])

    gen.emit_move(xtoy.re, dst_re)
    gen.emit_move(xtoy.im, dst_im)
    gen.emit_label(done)

    return ComplexArg(dst_re, dst_im)


def lt_cc_b(gen, t, srcs):
    # compare real parts only
    return gen.emit_binop(t.op, reals(srcs), Bool)


# these comparisons implemented the same way
lte_cc_b = gt_cc_b = gte_cc_b = lt_cc_b


def eq_cc_b(gen, t, srcs):
    # compare 2 complex numbers for equality
    d1 = gen.emit_binop(t.op, reals(srcs), Bool)
    d2 = gen.emit_binop(t.op, imags(srcs), Bool)
    dst = gen.emit_binop("&&", [d1, d2], Bool)
    return dst


def noteq_cc_b(gen, t, srcs):
    # compare 2 complex numbers for inequality
    d1 = gen.emit_binop(t.op, reals(srcs), Bool)
    d2 = gen.emit_binop(t.op, imags(srcs), Bool)
    dst = gen.emit_binop("||", [d1, d2], Bool)
    return dst

# sqr = square(x) = x*x


def sqr_c_c(gen, t, srcs):
    # sqr(a+ib) = a2 - b2 + i(2ab)
    src = srcs[0]
    a2 = gen.emit_binop('*', [src.re, src.re], Float)
    b2 = gen.emit_binop('*', [src.im, src.im], Float)
    ab = gen.emit_binop('*', [src.re, src.im], Float)
    dst = ComplexArg(
        gen.emit_binop('-', [a2, b2], Float),
        gen.emit_binop('*', [ConstFloatArg(2.0), ab], Float))
    return dst


def sqr_f_f(gen, t, srcs):
    return gen.emit_binop('*', [srcs[0], srcs[0]], Float)


sqr_i_i = sqr_f_f


def conj_c_c(gen, t, srcs):
    # conj (a+ib) = a-ib
    b = gen.emit_binop('-', [ConstFloatArg(0.0), srcs[0].im], Float)
    return ComplexArg(srcs[0].re, b)


def flip_c_c(gen, t, srcs):
    # flip(a+ib) = b+ia
    return ComplexArg(srcs[0].im, srcs[0].re)


def imag_c_f(gen, t, srcs):
    return srcs[0].im


def imag2_c_f(gen, t, srcs):
    return gen.emit_binop('*', [srcs[0].im, srcs[0].im], Float)


def real_c_f(gen, t, srcs):
    return srcs[0].re


def real2_c_f(gen, t, srcs):
    return gen.emit_binop('*', [srcs[0].re, srcs[0].re], Float)


def real_h_f(gen, t, srcs):
    return srcs[0].parts[0]


def imag_h_f(gen, t, srcs):
    return srcs[0].parts[1]


def hyper_j_h_f(gen, t, srcs):
    return srcs[0].parts[2]


def hyper_k_h_f(gen, t, srcs):
    return srcs[0].parts[3]


red_C_f = real_h_f
green_C_f = imag_h_f
blue_C_f = hyper_j_h_f
alpha_C_f = hyper_k_h_f


def hyper_ri_h_c(gen, t, srcs):
    return ComplexArg(srcs[0].parts[0], srcs[0].parts[1])


def hyper_jk_h_c(gen, t, srcs):
    return ComplexArg(srcs[0].parts[2], srcs[0].parts[3])


def ident_i_i(gen, t, srcs):
    return srcs[0]


ident_f_f = ident_c_c = ident_h_h = ident_b_b = ident_i_i

bool_b_b = ident_b_b
int_i_i = ident_i_i
float_f_f = ident_f_f


def recip_f_f(gen, t, srcs):
    # reciprocal
    return gen.emit_binop('/', [ConstFloatArg(1.0), srcs[0]], Float)


def recip_c_c(gen, t, srcs):
    return div_cc_c(gen, None,
                    [ComplexArg(ConstFloatArg(1.0), ConstFloatArg(0.0)), srcs[0]])


def recip_h_h(gen, t, srcs):
    src = srcs[0]
    re = src.parts[0]
    i = src.parts[1]
    j = src.parts[2]
    k = src.parts[3]

    # det = ((re-k)^2 + (i+j)^2)*((re+k)^2 + (i-j)^2)
    re_m_k = gen.emit_binop('-', [re, k], Float)
    re_p_k = gen.emit_binop('+', [re, k], Float)
    i_m_j = gen.emit_binop('-', [i, j], Float)
    i_p_j = gen.emit_binop('+', [i, j], Float)

    det = gen.emit_binop('*', [
        gen.emit_binop('+', [
            gen.emit_binop('*', [re_m_k, re_m_k], Float),
            gen.emit_binop('*', [i_p_j, i_p_j], Float),
        ], Float),
        gen.emit_binop('+', [
            gen.emit_binop('*', [re_p_k, re_p_k], Float),
            gen.emit_binop('*', [i_m_j, i_m_j], Float),
        ], Float),
    ], Float)

    # could check if det == 0 but just allow division by zero instead

    # |h|
    mod = cmag_h_f(gen, t, srcs)

    # 2 * re*k - i*j
    re_k_minus_ij = gen.emit_binop('-', [
        gen.emit_binop('*', [re, k], Float),
        gen.emit_binop('*', [i, j], Float)], Float)
    re_k_minus_ij = gen.emit_binop('+', [re_k_minus_ij, re_k_minus_ij], Float)

    return HyperArg(
        gen.emit_binop('/', [
            gen.emit_binop('-', [
                gen.emit_binop('*', [re, mod], Float),
                gen.emit_binop('*', [k, re_k_minus_ij], Float)], Float),
            det], Float),
        gen.emit_binop('/', [
            gen.emit_binop('-', [
                neg_f_f(gen, t, [gen.emit_binop('*', [i, mod], Float)]),
                gen.emit_binop('*', [j, re_k_minus_ij], Float)], Float),
            det], Float),
        gen.emit_binop('/', [
            gen.emit_binop('-', [
                neg_f_f(gen, t, [gen.emit_binop('*', [j, mod], Float)]),
                gen.emit_binop('*', [i, re_k_minus_ij], Float)], Float),
            det], Float),
        gen.emit_binop('/', [
            gen.emit_binop('-', [
                gen.emit_binop('*', [k, mod], Float),
                gen.emit_binop('*', [re, re_k_minus_ij], Float)], Float),
            det], Float))


def abs_i_i(gen, t, srcs):
    return gen.emit_func('abs', srcs, Int)


def abs_f_f(gen, t, srcs):
    return gen.emit_func('fabs', srcs, Float)


def abs_c_c(gen, t, srcs):
    return ComplexArg(
        abs_f_f(gen, t, [srcs[0].re]), abs_f_f(gen, t, [srcs[0].im]))


def cabs_c_f(gen, t, srcs):
    # FIXME: per std_complex.h,should divide numbers first to avoid overflow
    return sqrt_f_f(gen, t, [cmag_c_f(gen, t, srcs)])


def sqrt_f_f(gen, t, srcs):
    return gen.emit_func('sqrt', srcs, Float)


def min2_c_f(gen, t, srcs):
    r2 = real2_c_f(gen, t, srcs)
    i2 = imag2_c_f(gen, t, srcs)
    real_larger = gen.symbols.newLabel()
    done = gen.symbols.newLabel()
    dst = gen.newTemp(Float)

    rgt = gen.emit_binop('>=', [r2, i2], Float)
    gen.emit_cjump(rgt, real_larger)

    # imag larger
    gen.emit_move(r2, dst)
    gen.emit_jump(done)

    gen.emit_label(real_larger)
    gen.emit_move(i2, dst)
    gen.emit_label(done)

    return dst


def min_ff_f(gen, t, srcs):
    l = srcs[0]
    r = srcs[1]
    right_larger = gen.symbols.newLabel()
    done = gen.symbols.newLabel()
    dst = gen.newTemp(Float)
    right_greater = gen.emit_binop('>', [r, l], Float)
    gen.emit_cjump(right_greater, right_larger)
    # l is larger
    gen.emit_move(r, dst)
    gen.emit_jump(done)
    gen.emit_label(right_larger)
    gen.emit_move(l, dst)
    gen.emit_label(done)
    return dst


def max_ff_f(gen, t, srcs):
    l = srcs[0]
    r = srcs[1]
    right_larger = gen.symbols.newLabel()
    done = gen.symbols.newLabel()
    dst = gen.newTemp(Float)
    right_greater = gen.emit_binop('>', [r, l], Float)
    gen.emit_cjump(right_greater, right_larger)
    # l is larger
    gen.emit_move(l, dst)
    gen.emit_jump(done)
    gen.emit_label(right_larger)
    gen.emit_move(r, dst)
    gen.emit_label(done)
    return dst


def max2_c_f(gen, t, srcs):
    r2 = real2_c_f(gen, t, srcs)
    i2 = imag2_c_f(gen, t, srcs)
    real_larger = gen.symbols.newLabel()
    done = gen.symbols.newLabel()
    dst = gen.newTemp(Float)

    rgt = gen.emit_binop('>=', [r2, i2], Float)
    gen.emit_cjump(rgt, real_larger)

    # imag larger
    gen.emit_move(i2, dst)
    gen.emit_jump(done)

    gen.emit_label(real_larger)
    gen.emit_move(r2, dst)
    gen.emit_label(done)

    return dst


def sqrt_c_c(gen, t, srcs):
    xnonzero = gen.symbols.newLabel()
    done = gen.symbols.newLabel()
    dst_re = gen.newTemp(Float)
    dst_im = gen.newTemp(Float)

    gen.emit_cjump(srcs[0].re, xnonzero)

    # only an imaginary part :
    # temp = sqrt(abs(z.im) / 2);
    # return (temp, __y < 0 ? -__temp : __temp);

    temp = sqrt_f_f(gen, t, [abs_f_f(gen, t, [
        gen.emit_binop('/', [srcs[0].im, ConstFloatArg(2.0)], Float)])])

    gen.emit_move(temp, dst_re)
    # y >= 0?
    ypos = gen.emit_binop('>=', [srcs[0].im, ConstFloatArg(0.0)], Float)
    ygtzero = gen.symbols.newLabel()
    gen.emit_cjump(ypos, ygtzero)

    nt = neg_f_f(gen, t, [temp])
    gen.emit_move(nt, temp)

    gen.emit_label(ygtzero)
    gen.emit_move(temp, dst_im)
    gen.emit_jump(done)

    gen.emit_label(xnonzero)
    # both real and imaginary

    # temp = sqrt(2 * (cabs(z) + abs(z.re)));
    # u = temp/2
    temp = sqrt_f_f(
        gen, t,
        [gen.emit_binop(
            '*',
            [ConstFloatArg(2.0),
             gen.emit_binop(
                 '+',
                 [cabs_c_f(gen, t, [srcs[0]]),
                  abs_f_f(gen, t, [srcs[0].re])],
                 Float)
             ],
            Float)
         ])
    u = gen.emit_binop('/', [temp, ConstFloatArg(2.0)], Float)

    # x > 0?
    xpos = gen.emit_binop('>', [srcs[0].re, ConstFloatArg(0.0)], Float)
    xgtzero = gen.symbols.newLabel()
    gen.emit_cjump(xpos, xgtzero)

    # x < 0:

    # x = abs(im)/temp
    gen.emit_move(gen.emit_binop(
        '/',
        [abs_f_f(gen, t, [srcs[0].im]), temp], Float), dst_re)

    # y < 0 ? -u : u
    ypos2 = gen.emit_binop('>=', [srcs[0].im, ConstFloatArg(0.0)], Float)
    ygtzero2 = gen.symbols.newLabel()
    gen.emit_cjump(ypos2, ygtzero2)
    gen.emit_move(neg_f_f(gen, t, [u]), dst_im)
    gen.emit_jump(done)
    gen.emit_label(ygtzero2)
    gen.emit_move(u, dst_im)
    gen.emit_jump(done)

    # x > 0:
    gen.emit_label(xgtzero)

    # (u, im/temp)
    gen.emit_move(u, dst_re)
    gen.emit_move(gen.emit_binop('/', [srcs[0].im, temp], Float), dst_im)

    gen.emit_label(done)

    return ComplexArg(dst_re, dst_im)


def sin_f_f(gen, t, srcs):
    return gen.emit_func('sin', srcs, Float)


def sin_c_c(gen, t, srcs):
    # sin(a+ib) = (sin(a) * cosh(b), cos(a) * sinh(b))
    a = srcs[0].re
    b = srcs[0].im
    re = gen.emit_binop(
        '*', [sin_f_f(gen, t, [a]), cosh_f_f(gen, t, [b])], Float)
    im = gen.emit_binop(
        '*', [cos_f_f(gen, t, [a]), sinh_f_f(gen, t, [b])], Float)
    return ComplexArg(re, im)


def cos_f_f(gen, t, srcs):
    return gen.emit_func('cos', srcs, Float)


def cos_c_c(gen, t, srcs):
    # cos(a+ib) = (cos(a) * cosh(b), -(sin(a) * sinh(b)))
    a = srcs[0].re
    b = srcs[0].im
    re = gen.emit_binop(
        '*', [cos_f_f(gen, t, [a]), cosh_f_f(gen, t, [b])], Float)
    im = gen.emit_binop(
        '*', [sin_f_f(gen, t, [a]), sinh_f_f(gen, t, [b])], Float)

    nim = gen.emit_func('-', [im], Float)
    return ComplexArg(re, nim)


def cosxx_c_c(gen, t, srcs):
    # cosxx(z) = (re(cos(z)), -im(cos(z)))
    return conj_c_c(gen, t, [cos_c_c(gen, t, srcs)])


def tan_f_f(gen, t, srcs):
    return gen.emit_func('tan', srcs, Float)


def tan_c_c(gen, t, srcs):
    # tan = sin/cos
    return div_cc_c(
        gen, t, [sin_c_c(gen, t, [srcs[0]]), cos_c_c(gen, t, [srcs[0]])])


def cotan_c_c(gen, t, srcs):
    return div_cc_c(
        gen, t, [cos_c_c(gen, t, [srcs[0]]), sin_c_c(gen, t, [srcs[0]])])


def cotan_f_f(gen, t, srcs):
    return gen.emit_binop('/', [gen.emit_func('cos', srcs, Float),
                                gen.emit_func('sin', srcs, Float)], Float)


def cotanh_f_f(gen, t, srcs):
    return gen.emit_binop('/', [gen.emit_func('cosh', srcs, Float),
                                gen.emit_func('sinh', srcs, Float)], Float)


def cotanh_c_c(gen, t, srcs):
    return div_cc_c(gen, t, [cosh_c_c(gen, t, [srcs[0]]),
                             sinh_c_c(gen, t, [srcs[0]])])


def cosh_f_f(gen, t, srcs):
    return gen.emit_func('cosh', srcs, Float)


def cosh_c_c(gen, t, srcs):
    # cosh(a+ib) = cosh(a)*cos(b) + i (sinh(a) * sin(b))
    a = [srcs[0].re]
    b = [srcs[0].im]
    re = gen.emit_binop('*', [cosh_f_f(gen, t, a), cos_f_f(gen, t, b)], Float)
    im = gen.emit_binop('*', [sinh_f_f(gen, t, a), sin_f_f(gen, t, b)], Float)
    return ComplexArg(re, im)


def sinh_f_f(gen, t, srcs):
    return gen.emit_func('sinh', srcs, Float)


def sinh_c_c(gen, t, srcs):
    # sinh(a+ib) = sinh(a)*cos(b) + i (cosh(a) * sin(b))
    a = [srcs[0].re]
    b = [srcs[0].im]
    re = gen.emit_binop('*', [sinh_f_f(gen, t, a), cos_f_f(gen, t, b)], Float)
    im = gen.emit_binop('*', [cosh_f_f(gen, t, a), sin_f_f(gen, t, b)], Float)
    return ComplexArg(re, im)


def tanh_f_f(gen, t, srcs):
    return gen.emit_func('tanh', srcs, Float)


def tanh_c_c(gen, t, srcs):
    # tanh = sinh / cosh
    return div_cc_c(
        gen, t, [sinh_c_c(gen, t, [srcs[0]]), cosh_c_c(gen, t, [srcs[0]])])


def asin_f_f(gen, t, srcs):
    return gen.emit_func('asin', srcs, Float)


def asin_c_c(gen, t, srcs):
    # asin(z) = -i * log(i*z + sqrt(1-z*z))

    one_minus_z2 = sub_cc_c(gen, t, [Constants.one, sqr_c_c(gen, t, srcs)])
    sq = sqrt_c_c(gen, t, [one_minus_z2])
    arg = add_cc_c(gen, t, [mul_cc_c(gen, t, [Constants.i, srcs[0]]), sq])

    l = log_c_c(gen, t, [arg])
    return mul_cc_c(gen, t, [Constants.minus_i, l])


def acos_f_f(gen, t, srcs):
    return gen.emit_func('acos', srcs, Float)


def acos_c_c(gen, t, srcs):
    # acos(z) = pi/2 - asin(z)
    pi_by_2 = ComplexArg(ConstFloatArg(math.pi / 2.0), ConstFloatArg(0.0))
    return sub_cc_c(gen, t, [pi_by_2, asin_c_c(gen, t, srcs)])


def atan_f_f(gen, t, srcs):
    return gen.emit_func('atan', srcs, Float)


def atan_c_c(gen, t, srcs):
    # atan(z) = i/2 * log(i+x/i-x)
    ratio = div_cc_c(gen, t, [add_cc_c(gen, t, [Constants.i, srcs[0]]),
                              sub_cc_c(gen, t, [Constants.i, srcs[0]])])
    return mul_cc_c(gen, t, [Constants.iby2, log_c_c(gen, t, [ratio])])


def trunc_f_i(gen, t, srcs):
    return gen.emit_func("(int)", srcs, Int)


def trunc_c_c(gen, t, srcs):
    return ComplexArg(trunc_f_i(gen, t, [srcs[0].re]),
                      trunc_f_i(gen, t, [srcs[0].im]))


def round_f_i(gen, t, srcs):
    return trunc_f_i(gen, t, [
        gen.emit_binop('+', [ConstFloatArg(0.5), srcs[0]], Float)])


def round_c_c(gen, t, srcs):
    return ComplexArg(round_f_i(gen, t, [srcs[0].re]),
                      round_f_i(gen, t, [srcs[0].im]))


def floor_f_i(gen, t, srcs):
    return gen.emit_func('floor', srcs, Int)


def floor_c_c(gen, t, srcs):
    return ComplexArg(floor_f_i(gen, t, [srcs[0].re]),
                      floor_f_i(gen, t, [srcs[0].im]))


def ceil_f_i(gen, t, srcs):
    return gen.emit_func('ceil', srcs, Int)


def ceil_c_c(gen, t, srcs):
    return ComplexArg(ceil_f_i(gen, t, [srcs[0].re]),
                      ceil_f_i(gen, t, [srcs[0].im]))


def zero_i_i(gen, t, srcs):
    return ConstIntArg(0)


def zero_f_f(gen, t, srcs):
    return ConstFloatArg(0.0)


def zero_c_c(gen, t, srcs):
    return ComplexArg(ConstFloatArg(0.0), ConstFloatArg(0.0))


def atan2_c_f(gen, t, srcs):
    return gen.emit_func2('atan2', [srcs[0].im, srcs[0].re], Float)


def asinh_f_f(gen, t, srcs):
    return gen.emit_func('asinh', srcs, Float)


def asinh_c_c(gen, t, srcs):
    # log(z + sqrt(z*z+1))
    sq = sqrt_c_c(
        gen, t, [
            add_cc_c(
                gen, t, [
                    Constants.one, sqr_c_c(
                        gen, t, srcs)])])
    return log_c_c(gen, t, [add_cc_c(gen, t, [srcs[0], sq])])


def acosh_f_f(gen, t, srcs):
    return gen.emit_func('acosh', srcs, Float)


def acosh_c_c(gen, t, srcs):
    # log(z + sqrt(z-1)*sqrt(z+1))
    sqzm1 = sqrt_c_c(gen, t, [sub_cc_c(gen, t, [srcs[0], Constants.one])])
    sqzp1 = sqrt_c_c(gen, t, [add_cc_c(gen, t, [srcs[0], Constants.one])])
    sum = add_cc_c(gen, t, [srcs[0], mul_cc_c(gen, t, [sqzm1, sqzp1])])
    return log_c_c(gen, t, [sum])


def atanh_f_f(gen, t, srcs):
    return gen.emit_func('atanh', srcs, Float)


def times_i(gen, t, srcs):
    # multiply by i = (-im,re)
    return ComplexArg(neg_f_f(gen, t, [srcs[0].im]), srcs[0].re)


def atanh_c_c(gen, t, srcs):
    # 1/2(log(1+z)-log(1-z))
    one_m_z = log_c_c(gen, t, [sub_cc_c(gen, t, [Constants.one, srcs[0]])])
    one_p_z = log_c_c(gen, t, [add_cc_c(gen, t, [Constants.one, srcs[0]])])
    return mul_cf_c(gen, t, [sub_cc_c(gen, t, [one_p_z, one_m_z]),
                             ConstFloatArg(0.5)])


def manhattanish_c_f(gen, t, srcs):
    return gen.emit_binop('+', [srcs[0].re, srcs[0].im], Float)


def manhattan_c_f(gen, t, srcs):
    return gen.emit_binop('+', [abs_f_f(gen, t, [srcs[0].re]),
                                abs_f_f(gen, t, [srcs[0].im])], Float)


def manhattanish2_c_f(gen, t, srcs):
    return sqr_f_f(
        gen, t, [gen.emit_binop('+', [sqr_f_f(gen, t, [srcs[0].re]),
                                      sqr_f_f(gen, t, [srcs[0].im])], Float)])


def make_hyper_func(f):
    # takes a function from C -> C and returns a function which does the same
    # thing for H -> H
    def genfunc(gen, t, srcs):
        src = srcs[0]
        ax = gen.emit_binop('-', [src.parts[0], src.parts[3]], Float)
        ay = gen.emit_binop('+', [src.parts[1], src.parts[2]], Float)
        bx = gen.emit_binop('+', [src.parts[0], src.parts[3]], Float)
        by = gen.emit_binop('-', [src.parts[1], src.parts[2]], Float)

        res_a = f(gen, t, [ComplexArg(ax, ay)])
        res_b = f(gen, t, [ComplexArg(bx, by)])

        outa = gen.emit_binop('+', [res_a.re, res_b.re], Float)
        outb = gen.emit_binop('+', [res_a.im, res_b.im], Float)
        outc = gen.emit_binop('-', [res_a.im, res_b.im], Float)
        outd = gen.emit_binop('-', [res_b.re, res_a.re], Float)
        # divide by 2
        return mul_hf_h(gen, t, [HyperArg(outa, outb, outc, outd),
                                 ConstFloatArg(0.5)])
    return genfunc


sin_h_h = make_hyper_func(sin_c_c)
cos_h_h = make_hyper_func(cos_c_c)
cosxx_h_h = make_hyper_func(cosxx_c_c)
tan_h_h = make_hyper_func(tan_c_c)
cotan_h_h = make_hyper_func(cotan_c_c)

sinh_h_h = make_hyper_func(sinh_c_c)
cosh_h_h = make_hyper_func(cosh_c_c)
tanh_h_h = make_hyper_func(tanh_c_c)
cotanh_h_h = make_hyper_func(cotanh_c_c)

asin_h_h = make_hyper_func(asin_c_c)
acos_h_h = make_hyper_func(acos_c_c)
atan_h_h = make_hyper_func(atan_c_c)

asinh_h_h = make_hyper_func(asinh_c_c)
acosh_h_h = make_hyper_func(acosh_c_c)
atanh_h_h = make_hyper_func(atanh_c_c)

log_h_h = make_hyper_func(log_c_c)
sqrt_h_h = make_hyper_func(sqrt_c_c)
exp_h_h = make_hyper_func(exp_c_c)

flip_h_h = make_hyper_func(flip_c_c)
conj_h_h = make_hyper_func(conj_c_c)
sqr_h_h = make_hyper_func(sqr_c_c)


def rgb_fff_C(gen, t, srcs):
    return ColorArg(srcs[0], srcs[1], srcs[2], ConstFloatArg(1.0))


def hsl_fff_C(gen, t, srcs):
    [d1, d2, d3] = gen.emit_func3_3("hsl_to_rgb", srcs, Float)
    return ColorArg(d1, d2, d3, ConstFloatArg(1.0))


def hsla_ffff_C(gen, t, srcs):
    [d1, d2, d3] = gen.emit_func3_3("hsl_to_rgb", srcs, Float)
    return ColorArg(d1, d2, d3, srcs[3])


def hsv_fff_C(gen, t, srcs):
    [d1, d2, d3] = gen.emit_func3_3("hsv_to_rgb", srcs, Float)
    return ColorArg(d1, d2, d3, ConstFloatArg(1.0))


def hue_C_f(gen, t, srcs):
    c = srcs[0]
    return gen.emit_func3("hue", [c.parts[0], c.parts[1], c.parts[2]], Float)


def sat_C_f(gen, t, srcs):
    c = srcs[0]
    return gen.emit_func3("sat", [c.parts[0], c.parts[1], c.parts[2]], Float)


def lum_C_f(gen, t, srcs):
    c = srcs[0]
    return gen.emit_func3("lum", [c.parts[0], c.parts[1], c.parts[2]], Float)


def rgba_ffff_C(gen, t, srcs):
    return ColorArg(srcs[0], srcs[1], srcs[2], srcs[3])


color_ffff_C = rgba_ffff_C


def blend_CCf_C(gen, t, srcs):
    (a, b, factor) = srcs
    one_m_factor = gen.emit_binop('-', [ConstFloatArg(1.0), factor], Float)
    return add_CC_C(
        gen, t,
        [mul_Cf_C(gen, t, [a, one_m_factor]),
         mul_Cf_C(gen, t, [b, factor])])


def compose_CCf_C(gen, t, srcs):
    (a, b, factor) = srcs

    # factor *alpha(b)
    factor = gen.emit_binop('*', [b.parts[3], factor], Float)
    blend = blend_CCf_C(gen, t, [a, b, factor])
    gen.emit_move(a.parts[3], blend.parts[3])
    return blend


def mergenormal_CC_C(gen, t, srcs):
    return srcs[1]


def mergemultiply_CC_C(gen, t, srcs):
    (a, b) = (srcs[0], srcs[1])
    return ColorArg(
        gen.emit_binop('*', [a.parts[0], b.parts[0]], Float),
        gen.emit_binop('*', [a.parts[1], b.parts[1]], Float),
        gen.emit_binop('*', [a.parts[2], b.parts[2]], Float),
        b.parts[3])


def gradient_Gf_C(gen, t, srcs):
    [d1, d2, d3] = gen.emit_func2_3("gradient", srcs, Float)
    # fixme get alpha from gradient
    return ColorArg(d1, d2, d3, ConstFloatArg(1.0))


def _image_Ic_C(gen, t, srcs):
    c = srcs[1]
    [d1, d2, d3] = gen.emit_func3_3(
        "image_lookup", [srcs[0], c.re, c.im], Float)
    return ColorArg(d1, d2, d3, ConstFloatArg(1.0))


def gradient_f_C(gen, t, srcs):
    grad = gen.get_gradient_var()
    return gradient_Gf_C(gen, t, [grad, srcs[0]])


def rand__c(gen, t, srcs):
    [d1, d2] = gen.emit_func0_2("fract_rand", srcs, Float)
    return ComplexArg(d1, d2)


def _alloc_avii_av(gen, t, srcs):
    d = gen.emit_func3("alloc_array1D", srcs, VoidArray)
    return d


def _alloc_aviii_av(gen, t, srcs):
    d = gen.emit_func_n(4, "alloc_array2D", srcs, VoidArray)
    return d


def _alloc_aviiii_av(gen, t, srcs):
    d = gen.emit_func_n(5, "alloc_array3D", srcs, VoidArray)
    return d


def _alloc_aviiiii_av(gen, t, srcs):
    d = gen.emit_func_n(6, "alloc_array4D", srcs, VoidArray)
    return d


def _read_lookup_aii_i(gen, t, srcs):
    d = gen.emit_func2("read_int_array_1D", srcs, Int)
    return d


def _read_lookup_afi_f(gen, t, srcs):
    d = gen.emit_func2("read_float_array_1D", srcs, Float)
    return d


def _read_lookup_aci_c(gen, t, srcs):
    # lookup pair of floats
    i1 = gen.emit_binop('*', [ConstIntArg(2), srcs[1]], Int)
    i2 = gen.emit_binop('+', [ConstIntArg(1), i1], Int)
    d1 = gen.emit_func2("read_float_array_1D", [srcs[0], i1], Float)
    d2 = gen.emit_func2("read_float_array_1D", [srcs[0], i2], Float)

    return ComplexArg(d1, d2)


def _read_lookup_aiii_i(gen, t, srcs):
    d = gen.emit_func3("read_int_array_2D", srcs, Int)
    return d


def _read_lookup_afii_f(gen, t, srcs):
    d = gen.emit_func3("read_float_array_2D", srcs, Float)
    return d


def _read_lookup_acii_c(gen, t, srcs):

    # lookup pair of floats
    i1 = gen.emit_binop('*', [ConstIntArg(2), srcs[2]], Int)
    i2 = gen.emit_binop('+', [ConstIntArg(1), i1], Int)
    d1 = gen.emit_func3("read_float_array_2D", [srcs[0], srcs[1], i1], Float)
    d2 = gen.emit_func3("read_float_array_2D", [srcs[0], srcs[1], i2], Float)

    return ComplexArg(d1, d2)


def _read_lookup_aiiii_i(gen, t, srcs):
    pass


def _read_lookup_aiiiii_i(gen, t, srcs):
    pass


def _read_lookup_afiii_f(gen, t, srcs):
    pass


def _read_lookup_afiiii_f(gen, t, srcs):
    pass


def _read_lookup_aciii_c(gen, t, srcs):
    pass


def _read_lookup_aciiii_c(gen, t, srcs):
    pass


def _write_lookup_aiii_b(gen, t, srcs):
    d = gen.emit_func3("write_int_array_1D", srcs, Int)
    return d


def _write_lookup_afif_b(gen, t, srcs):
    d = gen.emit_func3("write_float_array_1D", srcs, Int)
    return d


def _write_lookup_acic_b(gen, t, srcs):
    # lookup pair of floats
    i1 = gen.emit_binop('*', [ConstIntArg(2), srcs[1]], Int)
    i2 = gen.emit_binop('+', [ConstIntArg(1), i1], Int)
    d1 = gen.emit_func3(
        "write_float_array_1D", [srcs[0], i1, srcs[2].re], Float)
    d2 = gen.emit_func3(
        "write_float_array_1D", [srcs[0], i2, srcs[2].im], Float)

    # CONSIDER: we ignore these anyway, but here we ignore
    # one return value 'extra thoroughly' which may need fixing
    # the proper return type would be a pair of bools
    return d1


def _write_lookup_aiiii_b(gen, t, srcs):
    d = gen.emit_func_n(4, "write_int_array_2D", srcs, Int)
    return d


def _write_lookup_afiif_b(gen, t, srcs):
    d = gen.emit_func_n(4, "write_float_array_2D", srcs, Int)
    return d


def _write_lookup_aciic_b(gen, t, srcs):
    # lookup pair of floats
    i1 = gen.emit_binop('*', [ConstIntArg(2), srcs[2]], Int)
    i2 = gen.emit_binop('+', [ConstIntArg(1), i1], Int)
    d1 = gen.emit_func_n(
        4, "write_float_array_2D", [srcs[0], srcs[1], i1, srcs[3].re], Float)
    d2 = gen.emit_func_n(
        4, "write_float_array_2D", [srcs[0], srcs[1], i2, srcs[3].im], Float)

    return d1


def _write_lookup_aiiiii_b(gen, t, srcs):
    pass


def _write_lookup_aiiiiii_b(gen, t, srcs):
    pass


def _write_lookup_afiiif_b(gen, t, srcs):
    pass


def _write_lookup_afiiiif_b(gen, t, srcs):
    pass


def _write_lookup_aciiic_b(gen, t, srcs):
    pass


def _write_lookup_aciiiic_b(gen, t, srcs):
    pass



# This is called by the main gnofract4d script

import os
import shutil

from fract4d_compiler import fc
from . import fractal, image


class T:
    def __init__(self, userConfig):
        self.userConfig = userConfig
        self.compiler = fc.Compiler(userConfig)
        self.f = fractal.T(self.compiler)

    def run(self, options):
        for path in options.extra_paths:
            self.compiler.add_func_path(path)

        if options.flags is not None:
            self.compiler.set_flags(options.flags)

        width = options.width or self.userConfig.getint("display", "width")
        height = options.height or self.userConfig.getint("display", "height")
        threads = options.threads or self.userConfig.getint(
            "general", "threads")

        if options.paramfile:
            self.load(options.paramfile)

        self.f.apply_options(options)
        self.f.antialias = options.antialias or \
            self.userConfig.getint("display", "antialias")

        outfile = self.compile(options)

        if options.buildonly is not None:
            self.buildonly(options, outfile)
            return

        if options.singlepoint:
            self.f.drawpoint()
        else:
            im = image.T(width, height)
            self.f.draw(im, threads)

        if options.save_filename:
            im.save(options.save_filename)

    def compile(self, options):
        if options.usebuilt is None:
            return self.f.compile()
        else:
            self.f.set_output_file(options.usebuilt)

    def buildonly(self, options, outfile):
        outdirname = os.path.dirname(options.buildonly)
        if len(outdirname) > 0:
            os.makedirs(outdirname)

        shutil.copy(outfile, options.buildonly)
        (base, ext) = os.path.splitext(outfile)
        cfile = base + ".c"
        shutil.copy(cfile, options.buildonly + ".c")

    def load(self, filename):
        with open(filename) as fh:
            self.f.loadFctFile(fh)



#!/usr/bin/python

import os.path
import math

from xml.sax import make_parser
from xml.sax.handler import ContentHandler

# interpolation type constants
INT_LINEAR = 0
INT_LOG = 1
INT_INVLOG = 2
INT_COS = 3


def getAttrOrDefault(attrs, name, default):
    x = attrs.get(name)
    if x is None:
        x = default
    return x


def getAttrOrElse(attrs, name):
    x = attrs.get(name)
    if x is None:
        raise ValueError(
            "Invalid file: Cannot find required attribute '%s'" % name)
    return x


class KeyFrame:
    def __init__(self, filename, duration, stop,
                 int_type, flags=(0, 0, 0, 0, 0, 0)):
        self.filename = filename
        self.duration = duration
        self.stop = stop
        self.int_type = int_type
        self.flags = flags

    def save(self, fh):
        fh.write(
            '\t\t<keyframe filename="%s" duration="%d" stop="%d" inttype="%d" ' % (self.filename, self.duration, self.stop, self.int_type))
        fh.write(
            'xy="%d" xz="%d" xw="%d" yz="%d" yw="%d" zw="%d"' % self.flags)
        fh.write('/>\n')

    @staticmethod
    def load_from_xml(attrs):
        kf = KeyFrame(
            getAttrOrElse(attrs, "filename"),
            int(getAttrOrElse(attrs, "duration")),
            int(getAttrOrElse(attrs, "stop")),
            int(getAttrOrElse(attrs, "inttype")),
            (int(getAttrOrDefault(attrs, "xy", 0)),
             int(getAttrOrDefault(attrs, "xz", 0)),
             int(getAttrOrDefault(attrs, "xw", 0)),
             int(getAttrOrDefault(attrs, "yz", 0)),
             int(getAttrOrDefault(attrs, "yw", 0)),
             int(getAttrOrDefault(attrs, "zw", 0))))
        return kf


class T:
    def __init__(self, compiler, userConfig):
        self.compiler = compiler
        self.userConfig = userConfig
        self.reset()

    def reset(self):
        self.avi_file = ""
        self.width = 640
        self.height = 480
        self.framerate = 25
        self.redblue = False
        # keyframes is a list of KeyFrame objects
        self.keyframes = []

    def get_fct_enabled(self):
        return self.userConfig.getboolean("director", "fct_enabled")

    def set_fct_enabled(self, fct_enabled):
        if fct_enabled:
            self.userConfig.set("director", "fct_enabled", "1")
        else:
            self.userConfig.set("director", "fct_enabled", "0")

    def get_fct_dir(self):
        return self.userConfig.get("director", "fct_dir")

    def set_fct_dir(self, dir):
        self.userConfig.set("director", "fct_dir", dir)

    def get_png_dir(self):
        return self.userConfig.get("director", "png_dir")

    def set_png_dir(self, dir):
        self.userConfig.set("director", "png_dir", dir)

    def get_avi_file(self):
        return self.avi_file

    def set_avi_file(self, file):
        if file is not None:
            self.avi_file = file
        else:
            self.avi_file = ""

    def get_width(self):
        return self.width

    def set_width(self, width):
        if width is not None:
            self.width = int(width)
        else:
            self.width = 640

    def get_height(self):
        return self.height

    def set_height(self, height):
        if height is not None:
            self.height = int(height)
        else:
            self.height = 480

    def get_framerate(self):
        return self.framerate

    def set_framerate(self, fr):
        if fr is not None:
            self.framerate = int(fr)
        else:
            self.framerate = 25

    def get_redblue(self):
        return self.redblue

    def set_redblue(self, rb):
        if rb is not None:
            if rb == 1:
                self.redblue = True
            elif rb == 0:
                self.redblue = False
                self.redblue = rb
        else:
            self.redblue = True

    def add_keyframe(self, filename, duration, stop, int_type, index=None):
        kf = KeyFrame(filename, duration, stop, int_type)
        if index is None:
            self.keyframes.append(kf)
        else:
            self.keyframes.insert(index, kf)

    def remove_keyframe(self, index):
        del self.keyframes[index:index + 1]

    def change_keyframe(self, index, duration, stop, int_type):
        if index < len(self.keyframes):
            kf = self.keyframes[index]
            kf.duration = duration
            kf.stop = stop
            kf.int_type = int_type

    def get_keyframe(self, index):
        return self.keyframes[index]

    def get_keyframe_filename(self, index):
        return self.keyframes[index].filename

    def get_keyframe_duration(self, index):
        return self.keyframes[index].duration

    def set_keyframe_duration(self, index, duration):
        if index < len(self.keyframes):
            self.keyframes[index].duration = duration

    def get_keyframe_stop(self, index):
        return self.keyframes[index].stop

    def set_keyframe_stop(self, index, stop):
        if index < len(self.keyframes):
            self.keyframes[index].stop = stop

    def get_keyframe_int(self, index):
        return self.keyframes[index].int_type

    def set_keyframe_int(self, index, int_type):
        if index < len(self.keyframes):
            self.keyframes[index].int_type = int_type

    def get_directions(self, index):
        return self.keyframes[index].flags

    def set_directions(self, index, drct):
        if index < len(self.keyframes):
            self.keyframes[index].flags = drct

    def keyframes_count(self):
        return len(self.keyframes)

    def __getstate__(self):
        odict = self.__dict__.copy()  # copy the dict since we change it
        # del odict['fh']              # remove filehandle entry
        return odict

    def __setstate__(self, dict):
        self.keyframes = []
        self.__dict__.update(dict)   # update attributes

    def load_animation(self, file):
        # save __dict__ if there was error
        odict = self.__dict__.copy()
        try:
            self.keyframes = []
            parser = make_parser()
            ah = AnimationHandler(self)
            parser.setContentHandler(ah)
            with open(file) as f:
                parser.parse(f)
        except Exception as err:
            # retrieve previous__dict__
            self.__setstate__(odict)
            raise

    def save_animation(self, file):
        with open(file, "w") as fh:
            fh.write('<?xml version="1.0"?>\n')
            fh.write("<animation>\n")
            fh.write('\t<keyframes>\n')
            for kf in self.keyframes:
                kf.save(fh)
            fh.write('\t</keyframes>\n')
            fh.write('\t<output filename="%s" framerate="%d" width="%d" height="%d" swap="%d"/>\n' %
                     (self.avi_file, self.framerate, self.width, self.height, self.redblue))
            fh.write("</animation>\n")

    # leftover from debugging purposes
    def pr(self):
        print(self.__dict__)

    def get_image_filename(self, n):
        "The filename of the image containing the Nth frame"
        return os.path.join(self.get_png_dir(), "image_%07d.png" % n)

    def get_fractal_filename(self, n):
        "The filename of the .fct file which generates the Nth frame"
        return os.path.join(self.get_fct_dir(), "file_%07d.fct" % n)

    def get_mu(self, int_type, x):
        if int_type == INT_LINEAR:
            mu = x
        elif int_type == INT_LOG:
            mu = math.log(x + 1, 2)
        elif int_type == INT_INVLOG:
            mu = (math.exp(x) - 1) / (math.e - 1)
        elif int_type == INT_COS:
            mu = (1 - math.cos(x * math.pi)) / 2
        else:
            raise ValueError("Unknown interpolation type %d" % int_type)
        return mu

    # create a list containing all the filenames of the frames
    def create_list(self):
        framelist = []
        current = 0
        for i in range(self.keyframes_count()):
            for j in range(
                    self.get_keyframe_stop(i)):  # output keyframe 'stop' times
                framelist.append(self.get_image_filename(current))
            current += 1

            if i < self.keyframes_count() - 1:
                # skip final frame which has no transitions following it
                for j in range(self.get_keyframe_duration(
                        i)):  # output all transition files minus keyframe
                    framelist.append(self.get_image_filename(current))
                    current += 1

        return framelist

    def get_keyframe_durations(self):
        durations = []
        for i in range(self.keyframes_count()):
            durations.append(self.get_keyframe_duration(i))

        return durations

    def get_total_frames(self):
        count = 0
        nframes = self.keyframes_count()
        for i in range(nframes):
            count += self.get_keyframe_stop(i)
            if i < nframes - 1:
                # don't count the last frame's duration
                count += self.get_keyframe_duration(i)
        return count


class AnimationHandler(ContentHandler):
    def __init__(self, animation):
        self.animation = animation

    def startElement(self, name, attrs):
        if name == "output":
            self.animation.set_avi_file(attrs.get("filename"))
            self.animation.set_framerate(attrs.get("framerate"))
            self.animation.set_width(attrs.get("width"))
            self.animation.set_height(attrs.get("height"))
            self.animation.set_redblue(int(attrs.get("swap")))
        elif name == "keyframe":
            kf = KeyFrame.load_from_xml(attrs)
            self.animation.keyframes.append(kf)
        return



# A widget which controls an angle of the fractal

import math
import gi

gi.require_version('Gdk', '3.0')
gi.require_version('Gtk', '3.0')
from gi.repository import Gtk, Gdk, GObject, Pango


class T(Gtk.DrawingArea):
    __gsignals__ = {
        'value-changed': (
            (GObject.SignalFlags.RUN_FIRST | GObject.SignalFlags.NO_RECURSE),
            None, (GObject.TYPE_DOUBLE,)),
        'value-slightly-changed': (
            (GObject.SignalFlags.RUN_FIRST | GObject.SignalFlags.NO_RECURSE),
            None, (GObject.TYPE_DOUBLE,))
    }
    two_pi = 2.0 * math.pi
    ptr_radius = 4

    def __init__(self, text):
        self.radius = 0
        self.text = text

        self.adjustment = Gtk.Adjustment(
            value=0,
            lower=-math.pi,
            upper=math.pi,
            step_increment=0.01,
            page_increment=0.01,
            page_size=0)

        self.old_value = self.adjustment.get_value()

        self.adjustment.connect('value-changed', self.onAdjustmentValueChanged)

        Gtk.DrawingArea.__init__(self)

        self.set_size_request(40, 40)

        self.set_events(
            Gdk.EventMask.BUTTON_RELEASE_MASK |
            Gdk.EventMask.BUTTON1_MOTION_MASK |
            Gdk.EventMask.POINTER_MOTION_HINT_MASK |
            Gdk.EventMask.ENTER_NOTIFY_MASK |
            Gdk.EventMask.LEAVE_NOTIFY_MASK |
            Gdk.EventMask.BUTTON_PRESS_MASK |
            Gdk.EventMask.EXPOSURE_MASK
        )

        self.notice_mouse = False
        self.connect('motion_notify_event', self.onMotionNotify)
        self.connect('button_release_event', self.onButtonRelease)
        self.connect('button_press_event', self.onButtonPress)
        self.connect('draw', self.onDraw)

    def set_value(self, val):
        val = math.fmod(val + math.pi, 2.0 * math.pi) - math.pi
        self.adjustment.set_value(val)
        self.queue_draw()

    def update_from_mouse(self, x, y):
        (w, h) = (self.get_allocated_width(), self.get_allocated_height())

        xc = w // 2
        yc = h // 2

        angle = self.angle_from_xy(x - xc, y - yc)
        self.adjustment.set_value(angle)

    def angle_from_xy(self, x, y):
        angle = math.atan2(y, x)
        #angle = math.fmod(angle,T.two_pi)
        return angle

    def onAdjustmentValueChanged(self, adjustment):
        val = adjustment.get_value()
        if val != self.old_value:
            self.queue_draw()
            self.emit('value-slightly-changed', val)

    def onMotionNotify(self, widget, event):
        if not self.notice_mouse:
            return
        self.update_from_mouse(event.x, event.y)

    def onButtonRelease(self, widget, event):
        if event.button == 1:
            self.notice_mouse = False
            current_value = self.adjustment.get_value()
            if self.old_value != current_value:
                self.old_value = current_value
                self.emit('value-changed', current_value)

        self.set_state(Gtk.StateType.NORMAL)

    def onButtonPress(self, widget, event):
        if event.button == 1:
            self.notice_mouse = True
            self.update_from_mouse(event.x, event.y)
            self.set_state(Gtk.StateType.ACTIVE)

    def get_current_angle(self):
        value = self.adjustment.get_value()
        return value

    def pointer_coords(self, radius, angle):
        s = math.sin(angle)
        c = math.cos(angle)

        ptr_xc = c * (radius - T.ptr_radius)
        ptr_yc = s * (radius - T.ptr_radius)
        return (ptr_xc, ptr_yc)

    def onDraw(self, widget, cairo_ctx):
        self.redraw_rect(widget, cairo_ctx)

    def redraw_rect(self, widget, cairo_ctx):
        style_ctx = widget.get_style_context()
        (w, h) = (widget.get_allocated_width(), widget.get_allocated_height())

        xc = w // 2
        yc = h // 2
        radius = min(w, h) // 2 - 1

        # text
        pango_ctx = widget.get_pango_context()
        layout = Pango.Layout(pango_ctx)
        layout.set_text(self.text, len(self.text))
        (text_width, text_height) = layout.get_pixel_size()
        Gtk.render_layout(style_ctx,
                          cairo_ctx,
                          xc - text_width // 2,
                          yc - text_height // 2,
                          layout)
        cairo_ctx.new_path()

        # outer circle
        cairo_ctx.arc(
            xc,
            yc,
            radius - 1,
            0,
            self.two_pi)
        cairo_ctx.set_line_width(1)
        cairo_ctx.stroke()

        # inner circle indicating current angle
        angle = self.get_current_angle()
        (ptr_xc, ptr_yc) = self.pointer_coords(radius - 2, angle)

        cairo_ctx.arc(
            int(xc + ptr_xc),
            int(yc + ptr_yc),
            T.ptr_radius,
            0,
            self.two_pi)
        cairo_ctx.fill()
        cairo_ctx.stroke()



# a class to hold a formula and all the parameter settings which go with it

import copy
import math
import random
import re
import io
import weakref

from fract4d_compiler import fracttypes, function
from . import gradient, image

# matches a complex number
cmplx_re = re.compile(r'\((.*?),(.*?)\)')
# matches a hypercomplex number
hyper_re = re.compile(r'\((.*?),(.*?),(.*?),(.*?)\)')


class T:
    def __init__(self, compiler, parent=None, prefix=None):
        self.compiler = compiler
        self.formula = None
        self.funcName = None
        self.funcFile = None
        self.params = []
        self.paramtypes = []
        self.dirty = False
        self.set_prefix(prefix)
        if parent:
            self.parent = weakref.ref(parent)
        else:
            self.parent = None

    def set_prefix(self, prefix):
        self.prefix = prefix
        if prefix is None:
            self.sectname = "function"
        elif prefix == "cf0":
            self.sectname = "outer"
        elif prefix == "cf1":
            self.sectname = "inner"
        elif prefix[0] == 't':
            self.sectname = "transform"
        else:
            raise ValueError("Unexpected prefix '%s' " % prefix)

    def set_initparams_from_formula(self, g):
        self.params = self.formula.symbols.default_params()
        self.paramtypes = self.formula.symbols.type_of_params()
        for i in range(len(self.paramtypes)):
            if self.paramtypes[i] == fracttypes.Gradient:
                self.params[i] = copy.copy(g)
            elif self.paramtypes[i] == fracttypes.Image:
                im = image.T(1, 1)
                #b = im.image_buffer()
                #b[0] = chr(216)
                #b[3] = chr(88)
                #b[4] = chr(192)
                #b[11] = chr(255)
                self.params[i] = im

    def reset_params(self):
        self.params = self.formula.symbols.default_params()
        self.paramtypes = self.formula.symbols.type_of_params()

    def copy_from(self, other):
        # copy the function overrides
        for name in self.func_names():
            self.set_named_func(name,
                                other.get_func_value(name))

        # copy the params
        self.params = [copy.copy(x) for x in other.params]

    def initvalue(self, name, warp_param=None):
        ord = self.order_of_name(name)
        type = self.formula.symbols[name].type

        if type == fracttypes.Complex:
            if warp_param == name:
                return "warp"
            else:
                return "(%.17f,%.17f)" % (
                    self.params[ord], self.params[ord + 1])
        elif type == fracttypes.Hyper or type == fracttypes.Color:
            return "(%.17f,%.17f,%.17f,%.17f)" % \
                   (self.params[ord], self.params[ord + 1],
                    self.params[ord + 2], self.params[ord + 3])
        elif type == fracttypes.Float:
            return "%.17f" % self.params[ord]
        elif type == fracttypes.Int:
            return "%d" % self.params[ord]
        elif type == fracttypes.Bool:
            return "%s" % int(self.params[ord])
        elif type == fracttypes.Gradient:
            return "[\n" + self.params[ord].serialize() + "]"
        elif type == fracttypes.Image:
            return "[\n" + self.params[ord].serialize() + "]"
        else:
            raise ValueError("Unknown type %s for param %s" % (type, name))

    def save_formula_params(self, file, warp_param=None, sectnum=None):
        if sectnum is None:
            print("[%s]" % self.sectname, file=file)
        else:
            print("[%s]=%d" % (self.sectname, sectnum), file=file)

        print("formulafile=%s" % self.funcFile, file=file)
        print("function=%s" % self.funcName, file=file)

        if(self.compiler.is_inline(self.funcFile, self.funcName)):
            contents = self.compiler.get_formula_text(
                self.funcFile, self.funcName)
            print("formula=[\n%s\n]" % contents, file=file)

        names = self.func_names()
        names.sort()
        for name in names:
            print("%s=%s" % (name, self.get_func_value(name)), file=file)
        names = self.param_names()
        names.sort()
        for name in names:
            print(
                "%s=%s" %
                (name,
                 self.initvalue(
                     name,
                     warp_param)),
                file=file)

        print("[endsection]", file=file)

    def func_names(self):
        return self.formula.symbols.func_names()

    def param_names(self):
        return self.formula.symbols.param_names()

    def params_of_type(self, type, readable=False):
        params = []
        op = self.formula.symbols.order_of_params()
        for name in list(op.keys()):
            if name != '__SIZE__':
                if self.formula.symbols[name].type == type:
                    if readable:
                        params.append(self.formula.symbols.demangle(name))
                    else:
                        params.append(name)
        return params

    def get_func_value(self, func_to_get):
        fname = self.formula.symbols.demangle(func_to_get)
        func = self.formula.symbols[fname]
        return func[0].cname

    def get_named_param_value(self, name):
        op = self.formula.symbols.order_of_params()
        ord = op.get(self.formula.symbols.mangled_name(name))
        return self.params[ord]

    def order_of_name(self, name):
        symbol_table = self.formula.symbols
        op = symbol_table.order_of_params()
        rn = symbol_table.mangled_name(name)
        ord = op.get(rn)
        if ord is None:
            # print "can't find %s (%s) in %s" % (name,rn,op)
            pass
        return ord

    def set_gradient(self, g):
        ord = self.order_of_name("@_gradient")
        self.params[ord] = g

    def try_set_named_item(self, name, val):
        # set the item if it exists, don't worry if it doesn't
        try:
            self.set_named_item(name, val)
        except KeyError:
            pass

    def text(self):
        "Return the text of this formula"
        return self.compiler.get_formula_text(
            self.funcFile, self.funcName)

    def set_named_item(self, name, val):
        sym = self.formula.symbols[name].first()
        if isinstance(sym, function.Func):
            self.set_named_func(name, val)
        else:
            self.set_named_param(name, val)

    def set_named_param(self, name, val):
        ord = self.order_of_name(name)
        if ord is None:
            # print "Ignoring unknown param %s" % name
            return

        t = self.formula.symbols[name].type
        if t == fracttypes.Complex:
            m = cmplx_re.match(val)
            if m is not None:
                re = float(m.group(1))
                im = float(m.group(2))
                if self.params[ord] != re:
                    self.params[ord] = re
                    self.changed()
                if self.params[ord + 1] != im:
                    self.params[ord + 1] = im
                    self.changed()
            elif val == "warp":
                self.parent().set_warp_param(name)
        elif t == fracttypes.Hyper or t == fracttypes.Color:
            m = hyper_re.match(val)
            if m is not None:
                for i in range(4):
                    val = float(m.group(i + 1))
                    if self.params[ord + i] != val:
                        self.params[ord + i] = val
                        self.changed()
        elif t == fracttypes.Float:
            newval = float(val)
            if self.params[ord] != newval:
                self.params[ord] = newval
                self.changed()
        elif t == fracttypes.Int:
            newval = int(val)
            if self.params[ord] != newval:
                self.params[ord] = newval
                self.changed()
        elif t == fracttypes.Bool:
            # don't use bool(val) - that makes "0" = True
            try:
                i = int(val)
                i = (i != 0)
            except ValueError:
                # an old release included a 'True' or 'False' string
                if val == "True":
                    i = 1
                else:
                    i = 0
            if self.params[ord] != i:
                self.params[ord] = i
                self.changed()
        elif t == fracttypes.Gradient:
            grad = gradient.Gradient()
            grad.load(io.StringIO(val))
            self.params[ord] = grad
            self.changed()
        elif t == fracttypes.Image:
            im = image.T(2, 2)
            self.params[ord] = im
            self.changed()
        else:
            raise ValueError("Unknown param type %s for %s" % (t, name))

    def set_named_func(self, func_to_set, val):
        fname = self.formula.symbols.demangle(func_to_set)
        func = self.formula.symbols.get(fname)
        return self.set_func(func[0], val)

    def zw_random(self, weirdness, size):
        factor = math.fabs(1.0 - math.log(size)) + 1.0
        return weirdness * (random.random() - 0.5) * 1.0 / factor

    def mutate(self, weirdness, size):
        for i in range(len(self.params)):
            if self.paramtypes[i] == fracttypes.Float:
                self.params[i] += self.zw_random(weirdness, size)
            elif self.paramtypes[i] == fracttypes.Int:
                # FIXME: need to be able to look up enum to find min/max
                pass
            elif self.paramtypes[i] == fracttypes.Bool:
                if random.random() < weirdness * 0.2:
                    self.params[i] = not self.params[i]

    def nudge_param(self, n, x, y):
        if x == 0 and y == 0:
            return False

        self.params[n] += (0.025 * x)
        self.params[n + 1] += (0.025 * y)
        self.changed()
        return True

    def set_param(self, n, val):
        # set the N'th param to val, after converting from a string
        t = self.paramtypes[n]

        if t == fracttypes.Float:
            val = float(val)
        elif t == fracttypes.Int:
            val = int(val)
        elif t == fracttypes.Bool:
            val = bool(val)
        else:
            raise ValueError("Unknown parameter type %s" % t)

        if self.params[n] != val:
            self.params[n] = val
            self.changed()
            return True
        return False

    def set_func(self, func, fname):
        if func.cname != fname:
            self.formula.symbols.set_std_func(func, fname)
            self.dirty = True
            self.changed()
            return True
        else:
            return False

    def changed(self):
        self.dirty = True
        if self.parent:
            self.parent().changed()

    def is_direct(self):
        return self.formula.is_direct()

    def set_formula(self, file, func, gradient):
        formula = self.compiler.get_formula(file, func, self.prefix)

        if formula is None:
            raise ValueError("no such formula: %s:%s" % (file, func))

        if formula.errors != []:
            raise ValueError("invalid formula '%s':\n%s" %
                             (func, "\n".join(formula.errors)))

        self.formula = formula
        self.funcName = func
        self.funcFile = file

        self.set_initparams_from_formula(gradient)

    def load_param_bag(self, bag):
        for (name, val) in list(bag.dict.items()):
            if name == "formulafile" or name == "function":
                pass
            else:
                self.try_set_named_item(name, val)

    def blend(self, other, ratio):
        # update in-place our settings so that they are a mixture with the
        # other
        if self.funcName != other.funcName or self.funcFile != other.funcFile:
            raise ValueError(
                "Cannot blend parameters between different formulas")

        for i in range(len(self.params)):
            (a, b) = (self.params[i], other.params[i])
            if self.paramtypes[i] == fracttypes.Float:
                self.params[i] = a * (1.0 - ratio) + b * ratio
            elif self.paramtypes[i] == fracttypes.Int:
                self.params[i] = int(a * (1.0 - ratio) + b * ratio)
            elif self.paramtypes[i] == fracttypes.Bool:
                if ratio >= 0.5:
                    self.params[i] = b
            else:
                # don't interpolate
                pass



# The top-level data structure behind the UI. Corresponds to the MainWindow

import re

from . import undo

# We eavesdrop on parameter-changed notifications from the gtkfractal
# and use those to compile the history


class Model:
    def __init__(self, f):
        # undo history
        self.seq = undo.Sequence()

        # used to prevent undo/redo from triggering new commands
        self.callback_in_progress = False
        self.f = f
        self.old_f = self.f.serialize()
        self.f.connect('parameters-changed', self.onParametersChanged)

    def block_callbacks(self):
        self.callback_in_progress = True

    def unblock_callbacks(self):
        self.callback_in_progress = False

    def callbacks_allowed(self):
        return not self.callback_in_progress

    def onParametersChanged(self, *args):
        if not self.callbacks_allowed():
            return

        # print "before:"
        # print self.dump_history()

        current = self.f.serialize()

        def set_fractal(val):
            self.f.freeze()
            # print "redo:", current[:100]
            self.f.deserialize(val)
            if self.f.thaw():
                self.block_callbacks()
                self.f.changed()
                self.unblock_callbacks()

        self.seq.do(set_fractal, current, set_fractal, self.old_f)
        self.old_f = current

        # print "after:"
        # print self.dump_history()

    def extract_x_from_dump(self, dump):
        "Get (x=number) from file"
        x_re = re.compile(r'x=.*')
        m = x_re.search(dump)
        if m:
            return m.group()
        return "eek"

    def dump_history(self):
        i = 0
        print("(redo,undo)")
        for he in self.seq.history:
            if i == self.seq.pos:
                print("-->", end=' ')
            else:
                print("   ", end=' ')
            print("(%s,%s)" %
                  (self.extract_x_from_dump(he.redo_data),
                   self.extract_x_from_dump(he.undo_data)))
            i += 1

    def undo(self):
        if self.seq.can_undo():
            # print "before undo:"
            # print self.dump_history()
            self.seq.undo()
            # print "after undo:"
            # print self.dump_history()

    def redo(self):
        if self.seq.can_redo():
            # print "before redo:"
            # print self.dump_history()
            self.seq.redo()
            # print "after redo:"
            # print self.dump_history()



# A widget which controls 2 linear dimensions of the fractal

import gi

gi.require_version('Gdk', '3.0')
gi.require_version('Gtk', '3.0')
from gi.repository import Gtk, Gdk, GObject, Pango


class T(Gtk.DrawingArea):
    __gsignals__ = {
        'value-changed': (
            (GObject.SignalFlags.RUN_FIRST | GObject.SignalFlags.NO_RECURSE),
            None, (GObject.TYPE_INT, GObject.TYPE_INT)),
        'value-slightly-changed': (
            (GObject.SignalFlags.RUN_FIRST | GObject.SignalFlags.NO_RECURSE),
            None, (GObject.TYPE_INT, GObject.TYPE_INT))
    }

    def __init__(self, text):
        self.button = 0
        self.radius = 0
        self.last_x = 0
        self.last_y = 0
        self.text = text
        Gtk.DrawingArea.__init__(self)

        self.set_size_request(53, 53)

        self.set_events(
            Gdk.EventMask.BUTTON_RELEASE_MASK |
            Gdk.EventMask.BUTTON1_MOTION_MASK |
            Gdk.EventMask.POINTER_MOTION_HINT_MASK |
            Gdk.EventMask.ENTER_NOTIFY_MASK |
            Gdk.EventMask.LEAVE_NOTIFY_MASK |
            Gdk.EventMask.BUTTON_PRESS_MASK |
            Gdk.EventMask.EXPOSURE_MASK
        )

        self.notice_mouse = False
        self.connect('motion_notify_event', self.onMotionNotify)
        self.connect('button_release_event', self.onButtonRelease)
        self.connect('button_press_event', self.onButtonPress)
        self.connect('draw', self.onDraw)

    def update_from_mouse(self, x, y):
        dx = self.last_x - x
        dy = self.last_y - y
        if dx or dy:
            self.emit('value-slightly-changed', dx, dy)
            self.last_x = x
            self.last_y = y

    def onMotionNotify(self, widget, event):
        if not self.notice_mouse:
            return
        self.update_from_mouse(event.x, event.y)

    def onButtonRelease(self, widget, event):
        if event.button == 1:
            self.notice_mouse = False
            (xc, yc) = (widget.get_allocated_width() //
                        2, widget.get_allocated_height() // 2)
            dx = xc - self.last_x
            dy = yc - self.last_y
            if dx or dy:
                self.emit('value-changed', dx, dy)

    def onButtonPress(self, widget, event):
        if event.button == 1:
            self.notice_mouse = True
            self.last_x = widget.get_allocated_width() / 2
            self.last_y = widget.get_allocated_height() / 2
            self.update_from_mouse(event.x, event.y)

    def onDraw(self, widget, cairo_ctx):
        self.redraw_rect(widget, cairo_ctx)

    def redraw_rect(self, widget, cairo_ctx):
        style_ctx = widget.get_style_context()
        (w, h) = (widget.get_allocated_width(), widget.get_allocated_height())
        Gtk.render_background(style_ctx, cairo_ctx, 0, 0, w - 1, h - 1)

        xc = w // 2
        yc = h // 2

        # Consider using gtk_render_arrow
        def triangle(points):
            cairo_ctx.move_to(*points[0])
            cairo_ctx.line_to(*points[1])
            cairo_ctx.line_to(*points[2])
            cairo_ctx.line_to(*points[0])
            cairo_ctx.fill()
            cairo_ctx.stroke()

        th = 8
        tw = 6

        # Triangle pointing left
        points = [
            (1, yc),
            (1 + th, yc - tw),
            (1 + th, yc + tw)]
        triangle(points)

        # pointing right
        points = [
            (w - 2, yc),
            (w - 2 - th, yc - tw),
            (w - 2 - th, yc + tw)]
        triangle(points)

        # pointing up
        points = [
            (xc, 1),
            (xc - tw, th),
            (xc + tw, th)]
        triangle(points)

        # pointing down
        points = [
            (xc, h - 2),
            (xc - tw, h - 2 - th),
            (xc + tw, h - 2 - th)]
        triangle(points)

        pango_ctx = widget.get_pango_context()
        layout = Pango.Layout(pango_ctx)

        try:
            drawtext = self.text
        except AttributeError:
            print("fourway has no text")
            return

        while True:
            layout.set_text(drawtext, len(drawtext))

            (text_width, text_height) = layout.get_pixel_size()
            # truncate text if it's too long
            if text_width < (w - th * 2) or len(drawtext) < 3:
                break
            drawtext = drawtext[:-1]

        Gtk.render_layout(
            style_ctx,
            cairo_ctx,
            xc - text_width // 2,
            yc - text_height // 2,
            layout)



#!/usr/bin/env python3
# Parser for UltraFractal + Fractint input files

from . import yacc, fractlexer, absyn

tokens = fractlexer.tokens

precedence = (
    ('left', 'COMMA'),
    ('right', 'ASSIGN'),
    ('left', 'BOOL_OR', 'BOOL_AND'),
    ('left', 'EQ', 'NEQ', 'LT', 'LTE', 'GT', 'GTE'),
    ('left', 'PLUS', 'MINUS'),
    ('left', 'TIMES', 'DIVIDE', 'MOD'),
    ('right', 'BOOL_NEG', 'UMINUS', 'POWER')
)


def formatError(t, i):
    if(isinstance(t[i], str)):
        e = [absyn.Error2(t[i], t.lineno(i))]
    else:
        e = [absyn.Error(t[i].type, t[i].value, t.lineno(i))]
    return e


def p_file(t):
    'file : formlist'
    t[0] = absyn.Formlist(t[1], t.lineno(1))


def p_formlist(t):
    'formlist : formula NEWLINE formlist'
    t[0] = [t[1]] + t[3]


def p_formlist_singleform(t):
    'formlist : formula'
    t[0] = [t[1]]


def p_formlist_empty(t):
    '''formlist : empty
       formlist : error'''
    t[0] = []


def p_formid(t):
    'formid : FORM_ID'
    t.set_lineno(0, t.lineno(1))
    t[0] = t[1]


def p_formula(t):
    'formula : formid formbody'
    #print("formula %s at %s" % (t[1], t.lineno(1)))
    t[0] = absyn.Formula(t[1], t[2], t.lineno(1))


def p_formula_err(t):
    'formula : error formid formbody'
    t[0] = absyn.Formula(t[2], t[3], t.lineno(2))


def p_formbody_2(t):
    'formbody : NEWLINE stmlist sectlist FORM_END'
    t[0] = [absyn.Stmlist("nameless", t[2], t.lineno(2))] + t[3]


def p_formbody_3(t):
    'formbody : NEWLINE stmlist FORM_END'
    t[0] = [absyn.Stmlist("nameless", t[2], t.lineno(2))]


def p_formbody_err(t):
    'formbody : error FORM_END'
    t[0] = formatError(t, 1)


def p_formbody_sectlist(t):
    'formbody : NEWLINE sectlist FORM_END'
    t[0] = t[2]


def p_sectlist_2(t):
    'sectlist : section sectlist'
    t[0] = listify(t[1]) + t[2]


def p_sectlist_section(t):
    'sectlist : section'
    t[0] = listify(t[1])


def p_section_parm(t):
    'section : SECT_PARMS parmlist'
    t[0] = absyn.Setlist(t[1], t[2], t.lineno(1))


def p_section_set(t):
    'section : SECT_SET setlist'
    t[0] = absyn.Setlist(t[1], t[2], t.lineno(1))


def p_setlist_set(t):
    'setlist : set'
    t[0] = listify(t[1])


def p_setlist_2(t):
    '''setlist : set NEWLINE setlist
       setlist : set COMMA setlist'''
    t[0] = listify(t[1]) + t[3]


def p_parmlist_parm(t):
    'parmlist : parm'
    t[0] = listify(t[1])


def p_parmlist_2(t):
    'parmlist : parm NEWLINE parmlist'
    t[0] = listify(t[1]) + t[3]


def p_parmlist_3(t):
    'parmlist : parm parmlist'
    t[0] = listify(t[1]) + t[2]


def p_parm_empty(t):
    'parm : empty'
    t[0] = t[1]

# the extras are to allow params which clash with types & reserved words


def p_parm_exp(t):
    '''parm : ID ASSIGN exp
       parm : TYPE ASSIGN exp
       parm : REPEAT ASSIGN exp'''
    t[0] = absyn.Set(absyn.ID(t[1], t.lineno(1)), t[3], t.lineno(2))


def p_parm_type(t):
    'parm : ID ASSIGN TYPE'
    t[0] = absyn.SetType(absyn.ID(t[1], t.lineno(1)), t[3], t.lineno(2))


def p_set_exp(t):
    'set : ID ASSIGN exp'
    t[0] = absyn.Set(absyn.ID(t[1], t.lineno(1)), t[3], t.lineno(2))


def p_set_type(t):
    'set : ID ASSIGN TYPE'
    t[0] = absyn.SetType(absyn.ID(t[1], t.lineno(1)), t[3], t.lineno(2))


def p_set_empty(t):
    'set : empty'
    t[0] = t[1]


def p_set_param(t):
    'set : PARAM ID setlist ENDPARAM'
    t[0] = absyn.Param(t[2], t[3], None, t.lineno(1))


def p_set_typed_param(t):
    'set : TYPE PARAM ID setlist ENDPARAM'
    t[0] = absyn.Param(t[3], t[4], t[1], t.lineno(2))


def p_set_func(t):
    'set : FUNC ID setlist ENDFUNC'
    t[0] = absyn.Func(t[2], t[3], "complex", t.lineno(1))


def p_set_typed_func(t):
    'set : TYPE FUNC ID setlist ENDFUNC'
    t[0] = absyn.Func(t[3], t[4], t[1], t.lineno(2))


def p_set_heading(t):
    'set : HEADING setlist ENDHEADING'
    t[0] = absyn.Heading(t[2], t.lineno(1))


def p_section_stm_2(t):
    'section : SECT_STM stmlist'
    t[0] = absyn.Stmlist(t[1], t[2], t.lineno(1))


def listify(stm):
    if stm.type == "empty":
        return []
    else:
        return [stm]


def p_stmlist_stm(t):
    'stmlist : stm'
    t[0] = listify(t[1])


def p_stmlist_2(t):
    '''stmlist : stm NEWLINE stmlist
       stmlist : stm COMMA stmlist'''
    t[0] = listify(t[1]) + t[3]


def p_stm_exp(t):
    'stm : exp'
    t[0] = t[1]


def p_stm_empty(t):
    'stm : empty'
    t[0] = t[1]


def p_empty(t):
    'empty :'
    t[0] = absyn.Empty(t.lineno(0))


def p_stm_decl_array(t):
    'stm : TYPE ID LARRAY arglist RARRAY'
    t[0] = absyn.DeclArray(t[1], t[2], t[4], t.lineno(3))


def p_stm_decl(t):
    'stm : TYPE ID'
    t[0] = absyn.Decl(t[1], t[2], t.lineno(2))


def p_stm_assign(t):
    'stm : TYPE ID ASSIGN stm'
    t[0] = absyn.Decl(t[1], t[2], t.lineno(2), t[4])


def p_stm_repeat(t):
    'stm : REPEAT stmlist UNTIL exp'
    t[0] = absyn.Repeat(t[2], t[4], t.lineno(1))


def p_stm_while(t):
    '''stm : WHILE exp NEWLINE stmlist ENDWHILE
       stm : WHILE exp COMMA stmlist ENDWHILE'''
    t[0] = absyn.While(t[2], t[4], t.lineno(1))


def p_stm_if(t):
    'stm : IF ifbody ENDIF'
    t[0] = t[2]


def p_ifbody(t):
    '''ifbody : exp NEWLINE stmlist
       ifbody : exp COMMA stmlist'''
    t[0] = absyn.If(t[1], t[3], [absyn.Empty(t.lineno(1))], t.lineno(1))


def p_ifbody_else(t):
    '''ifbody : exp NEWLINE stmlist ELSE stmlist
       ifbody : exp COMMA stmlist ELSE stmlist'''
    t[0] = absyn.If(t[1], t[3], t[5], t.lineno(1))


def p_ifbody_elseif(t):
    '''ifbody : exp NEWLINE stmlist ELSEIF ifbody
       ifbody : exp COMMA stmlist ELSEIF ifbody'''
    t[0] = absyn.If(t[1], t[3], [t[5]], t.lineno(1))


def p_exp_binop(t):
    '''exp : exp PLUS exp
       exp : exp MINUS exp
       exp : exp TIMES exp
       exp : exp DIVIDE exp
       exp : exp MOD exp
       exp : exp POWER exp
       exp : exp BOOL_OR exp
       exp : exp BOOL_AND exp
       exp : exp EQ exp
       exp : exp NEQ exp
       exp : exp LT exp
       exp : exp LTE exp
       exp : exp GT exp
       exp : exp GTE exp
       '''
    t[0] = absyn.Binop(t[2], t[1], t[3], t.lineno(2))


def p_exp_assign(t):
    'exp : lval ASSIGN exp'
    t[0] = absyn.Assign(t[1], t[3], t.lineno(2))


def p_lval_id_array(t):
    'lval : ID LARRAY arglist RARRAY'
    t[0] = absyn.ArrayLookup(t[1], t[3], t.lineno(1))


def p_lval_id(t):
    'lval : ID'
    t[0] = absyn.ID(t[1], t.lineno(1))


def p_lval_funcall(t):
    'lval : ID LPAREN arglist RPAREN'
    t[0] = absyn.Funcall(t[1], t[3], t.lineno(1))


def p_exp_uminus(t):
    'exp : MINUS exp %prec UMINUS'
    t[0] = absyn.Negate(t[2], t.lineno(1))

# unary plus is a no-op


def p_exp_uplus(t):
    'exp : PLUS exp %prec UMINUS'
    t[0] = t[2]


def p_exp_mag(t):
    'exp : MAG exp MAG'
    t[0] = absyn.Mag(t[2], t.lineno(1))


def p_exp_neg(t):
    'exp : BOOL_NEG exp'
    t[0] = absyn.Not(t[2], t.lineno(1))


def p_exp_num(t):
    'exp : NUMBER'
    t[0] = absyn.Number(t[1], t.lineno(1))


def p_exp_boolconst(t):
    'exp : CONST'
    t[0] = absyn.Const(t[1], t.lineno(1))


def p_exp_string(t):
    'exp : STRING stringlist'
    t[0] = absyn.String(t[1], t[2], t.lineno(1))


def p_stringlist_string(t):
    'stringlist : STRING stringlist'
    t[0] = [absyn.String(t[1], None, t.lineno(1))] + t[2]


def p_stringlist_empty(t):
    'stringlist : empty'
    t[0] = []


def p_exp_lval(t):
    'exp : lval'
    t[0] = t[1]


def p_exp_complex(t):
    'exp : LPAREN exp COMMA exp RPAREN'
    t[0] = absyn.Binop("complex", t[2], t[4], t.lineno(1))


def p_exp_hyper(t):
    'exp : LPAREN exp COMMA exp COMMA exp COMMA exp RPAREN'
    t[0] = absyn.Funcall("hyper", [t[2], t[4], t[6], t[8]], t.lineno(1))


def p_exp_ctor(t):
    'exp : TYPE LPAREN arglist RPAREN'
    t[0] = absyn.Funcall(t[1], t[3], t.lineno(1))


def p_exp_complex_i(t):
    'exp : COMPLEX'
    ln = t.lineno(1)
    t[0] = absyn.Binop(
        "complex", absyn.Number(
            "0.0", ln), absyn.Number(
            t[1], ln), ln)


def p_exp_funcall_noargs(t):
    'exp : ID LPAREN RPAREN'
    t[0] = absyn.Funcall(t[1], None, t.lineno(1))


def p_exp_parexp(t):
    'exp : LPAREN exp RPAREN'
    t[0] = t[2]


def p_arglist_exp(t):
    'arglist : exp'
    t[0] = [t[1]]


def p_arglist_2(t):
    'arglist : arglist COMMA arglist'
    t[0] = t[1] + t[3]

# Error rule for syntax errors outside a formula


def p_error(t):
    # print "error ",t
    pass


parser = yacc.yacc()

# debugging
if __name__ == '__main__':  # pragma: no cover
    import sys

    for arg in sys.argv[1:]:
        with open(arg, "r") as f:
            s = f.read()  # read in a whole file
        result = yacc.parse(s)
        print(result.pretty())

    if len(sys.argv) == 1:
        while True:
            try:
                s = input('calc > ')
            except EOFError:
                break
            if not s:
                continue
            if s[0] == '#':
                with open(s[1:], "r") as f:
                    s = f.read()  # read in a whole file
                print(s)
            else:
                s += "\n"
            result = yacc.parse(s)
            print("result", result)
            print(result.pretty())
